/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "quakedef.h"
#include "gl_modrdm.h"

entity_t	r_worldentity;

qboolean	r_cache_thrash;		// compatability

vec3_t		modelorg, r_entorigin;
entity_t	*currententity;

int			r_visframecount;	// bumped when going to a new PVS
int			r_framecount;		// used for dlight push checking

mplane_t	frustum[4];

int			c_brush_polys, c_alias_polys;

qboolean	envmap;				// true during envmap command capture 

int			currenttexture = -1;		// to avoid unnecessary texture sets

int			cnttextures[2] = {-1, -1};     // cached

int			particletexture;	// little dot for particles
int			playertextures;		// up to 16 color translated skins

int			mirrortexturenum;	// quake texturenum, not gltexturenum
qboolean	mirror;
mplane_t	*mirror_plane;

float g_sunPos[3]; //rww
qboolean g_hasSun = false; //rww

extern float g_glClearColor[4]; //rww
extern float g_beginLodCutoff; //rww
extern float g_endLodCutoff; //rww

extern unsigned int g_chaincount; //rww

extern cvar_t gl_twoSidedStencil; //rww

void GL_RenderSurfSprites(void);

//
// view origin
//
vec3_t	vup;
vec3_t	vpn;
vec3_t	vright;
vec3_t	r_origin;

float	r_world_matrix[16];
float	r_base_world_matrix[16];

float g_projectionMatrix[16]; //rww
viewFrustum_t g_curFrustum; //rww

float g_lightExposureFactor = 0.0f; //rww

surfSpriteDraw_t g_surfSpriteDraws[MAX_SURF_SPRITE_DRAWS];
int g_numSurfSpriteDraws = 0;

//
// screen size info
//
refdef_t	r_refdef;

mleaf_t		*r_viewleaf, *r_oldviewleaf;

texture_t	*r_notexture_mip;

int		d_lightstylevalue[256];	// 8.8 fraction of base light value


void R_MarkLeaves (void);

cvar_t	r_norefresh = {"r_norefresh","0"};
cvar_t	r_drawentities = {"r_drawentities","1"};
cvar_t	r_drawviewmodel = {"r_drawviewmodel","1"};
cvar_t	r_speeds = {"r_speeds","0"};
cvar_t	r_fullbright = {"r_fullbright","0"};
cvar_t	r_lightmap = {"r_lightmap","0"};
cvar_t	r_shadows = {"r_shadows","0"};
cvar_t	r_mirroralpha = {"r_mirroralpha","1"};
cvar_t	r_wateralpha = {"r_wateralpha","1.0"};
cvar_t	r_fancywater = {"r_fancywater","1"}; //rww
cvar_t	r_dynamic = {"r_dynamic","1"};
cvar_t	r_novis = {"r_novis","0"};
cvar_t	r_netgraph = {"r_netgraph","0"};

cvar_t	gl_clear = {"gl_clear","1"}; //rww - default 1
cvar_t	gl_cull = {"gl_cull","1"};
cvar_t	gl_texsort = {"gl_texsort","1"};
cvar_t	gl_smoothmodels = {"gl_smoothmodels","1"};
cvar_t	gl_affinemodels = {"gl_affinemodels","0"};
cvar_t	gl_polyblend = {"gl_polyblend","1"};
cvar_t	gl_flashblend = {"gl_flashblend","0"}; //rww - default 0
cvar_t	gl_playermip = {"gl_playermip","0"};
cvar_t	gl_nocolors = {"gl_nocolors","0"};
cvar_t	gl_keeptjunctions = {"gl_keeptjunctions","1"};
cvar_t	gl_reporttjunctions = {"gl_reporttjunctions","0"};
cvar_t	gl_finish = {"gl_finish","0"};
cvar_t	gl_fognear = {"gl_fognear","3000"}; //rww
cvar_t	gl_fogfar = {"gl_fogfar","4096"}; //rww
cvar_t	gl_fog_r = {"gl_fog_r","0.2"}; //rww //0.7
cvar_t	gl_fog_g = {"gl_fog_g","0.2"}; //rww //0.7
cvar_t	gl_fog_b = {"gl_fog_b","1.0"}; //rww //0.6

extern	cvar_t	gl_ztrick;
extern	cvar_t	scr_fov;
/*
=================
R_CullBox

Returns true if the box is completely outside the frustom
=================
*/
qboolean R_CullBox (vec3_t mins, vec3_t maxs)
{
	int		i;
	float	side;

	for (i=0 ; i<4 ; i++)
		if (BoxOnPlaneSide (mins, maxs, &frustum[i]) == 2)
			return true;

	/*
	if (BoxInFrustum(&g_curFrustum, mins, maxs) == 0)
	{ //rww - more aggressive check if simple one fails
		return true;
	}
	*/
	//rww
	side = g_curFrustum[5][0]*mins[0]+
	g_curFrustum[5][1]*mins[1]+
	g_curFrustum[5][2]*mins[2]+
	g_curFrustum[5][3];
	if (side <= 0.0f)
	{
		side = g_curFrustum[5][0]*maxs[0]+
		g_curFrustum[5][1]*maxs[1]+
		g_curFrustum[5][2]*maxs[2]+
		g_curFrustum[5][3];
		if (side <= 0.0f)
		{ //both points are outside of the far clipping plane
			float boundsPoints[8][3];
			GetBoundsPoints(mins, maxs, (float *)boundsPoints, NULL);
			for (i = 1; i < 7; i++)
			{
				side = g_curFrustum[5][0]*boundsPoints[i][0]+
				g_curFrustum[5][1]*boundsPoints[i][1]+
				g_curFrustum[5][2]*boundsPoints[i][2]+
				g_curFrustum[5][3];
				if (side > 0.0f)
				{ //oh well, this point wasn't in back
					return false;
				}
			}
			return true;
		}
	}

	return false;
}

//rww
qboolean GL_FancyWater(void)
{
	return (g_hasSun && g_canUseTexRect && g_canUseTexFP && gl_normalMaps.value && r_fancywater.value == 1.0f);
}

//rww
void GL_DebugDrawLineMinsMaxs(float *mins, float *maxs, float *color)
{
	float v[3];
	float c[3] = {1.0f, 0.0f, 0.0f};
	if (!color)
	{
		color = c;
	}

	glDisable(GL_TEXTURE_2D);
	glColor3fv(color);
	glBegin(GL_LINES);
		//lower
		v[0] = mins[0];
		v[1] = mins[1];
		v[2] = mins[2];
		glVertex3fv(v);

		v[0] = maxs[0];
		v[1] = mins[1];
		v[2] = mins[2];
		glVertex3fv(v);

		glVertex3fv(v);
		v[0] = maxs[0];
		v[1] = maxs[1];
		v[2] = mins[2];
		glVertex3fv(v);

		glVertex3fv(v);
		v[0] = mins[0];
		v[1] = maxs[1];
		v[2] = mins[2];
		glVertex3fv(v);

		glVertex3fv(v);
		v[0] = mins[0];
		v[1] = mins[1];
		v[2] = mins[2];
		glVertex3fv(v);
	glEnd();

	glBegin(GL_LINES);
		//upper
		v[0] = mins[0];
		v[1] = mins[1];
		v[2] = maxs[2];
		glVertex3fv(v);

		v[0] = maxs[0];
		v[1] = mins[1];
		v[2] = maxs[2];
		glVertex3fv(v);

		glVertex3fv(v);
		v[0] = maxs[0];
		v[1] = maxs[1];
		v[2] = maxs[2];
		glVertex3fv(v);

		glVertex3fv(v);
		v[0] = mins[0];
		v[1] = maxs[1];
		v[2] = maxs[2];
		glVertex3fv(v);

		glVertex3fv(v);
		v[0] = mins[0];
		v[1] = mins[1];
		v[2] = maxs[2];
		glVertex3fv(v);
	glEnd();

	glBegin(GL_LINES);
		//connect upper-lower
		v[0] = mins[0];
		v[1] = mins[1];
		v[2] = maxs[2];
		glVertex3fv(v);

		v[0] = mins[0];
		v[1] = mins[1];
		v[2] = mins[2];
		glVertex3fv(v);

		v[0] = maxs[0];
		v[1] = maxs[1];
		v[2] = maxs[2];
		glVertex3fv(v);

		v[0] = maxs[0];
		v[1] = maxs[1];
		v[2] = mins[2];
		glVertex3fv(v);

		v[0] = mins[0];
		v[1] = maxs[1];
		v[2] = maxs[2];
		glVertex3fv(v);

		v[0] = mins[0];
		v[1] = maxs[1];
		v[2] = mins[2];
		glVertex3fv(v);

		v[0] = maxs[0];
		v[1] = mins[1];
		v[2] = maxs[2];
		glVertex3fv(v);

		v[0] = maxs[0];
		v[1] = mins[1];
		v[2] = mins[2];
		glVertex3fv(v);
	glEnd();

	glColor3f(1.0f, 1.0f, 1.0f);
	glEnable(GL_TEXTURE_2D);
}

void R_RotateForEntity (entity_t *e)
{
    glTranslatef (e->origin[0],  e->origin[1],  e->origin[2]);

    glRotatef (e->angles[1],  0, 0, 1);
    glRotatef (-e->angles[0],  0, 1, 0);
	//ZOID: fixed z angle
    glRotatef (e->angles[2],  1, 0, 0);
}

//rww
extern int window_width, window_height;
extern int g_glFBOBloomTex;
extern unsigned int g_glFBOBloomDepth;
extern unsigned int g_glFBOBloomHandle;
extern int g_glFBOBloomBlur;
extern unsigned int g_glFBOBloomBlurHandle;
extern int g_bloomBufWidth;
extern int g_bloomBufHeight;
extern qboolean g_bloomLowQuality;
qboolean g_inFBORender = false;

//rww
static __inline void GL_BeginFBORender(unsigned int h, int width, int height)
{
	g_inFBORender = true;
	g_glExt.glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, h);
	glPushAttrib(GL_VIEWPORT_BIT);
	if (width == -1)
	{
		width = g_bloomBufWidth;
	}
	if (height == -1)
	{
		height = g_bloomBufHeight;
	}
	glViewport(0, 0, width, height);
}

//rww
static __inline void GL_EndFBORender(void)
{
	glPopAttrib();
	g_glExt.glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	g_inFBORender = false;
}

//rww
void GL_CapScreenTex(void)
{
	if (!g_canUseTexRect)
	{
		return;
	}

	glDisable(GL_TEXTURE_2D);
	glEnable(GL_TEXTURE_RECTANGLE_NV); 
	glBindTexture(GL_TEXTURE_RECTANGLE_NV, screen_texture);
	glCopyTexSubImage2D(GL_TEXTURE_RECTANGLE_NV, 0, 0, 0, 0, 0, window_width, window_height); 
	glDisable(GL_TEXTURE_RECTANGLE_NV);
	glEnable(GL_TEXTURE_2D);   
}

//rww
void GL_FullscreenQuad(int texture, int w, int h, qboolean isRect)
{
	float texw = 1.0f;
	float texh = 1.0f;
	if (isRect)
	{
		texw = (float)w;
		texh = (float)h;
	}

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, w, h, 0, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glDisable(GL_CULL_FACE);
	glDepthMask(GL_FALSE);
	glDisable(GL_DEPTH_TEST);
	if (texture != -1)
	{
		if (isRect)
		{
			glDisable(GL_TEXTURE_2D);
			glEnable(GL_TEXTURE_RECTANGLE_NV);
			glBindTexture(GL_TEXTURE_RECTANGLE_NV, texture);
		}
		else
		{
			glBindTexture(GL_TEXTURE_2D, texture);
		}
	}
	else
	{
		glDisable(GL_TEXTURE_2D);
	}
	glBegin(GL_QUADS);
		glTexCoord2f(0.0f, texh); 
		glVertex2f(0.0f, 0.0f);
		glTexCoord2f(0.0f, 0.0f);
		glVertex2f(0.0f, (float)h);
		glTexCoord2f(texw, 0.0f);
		glVertex2f((float)w, (float)h);
		glTexCoord2f(texw, texh);
		glVertex2f((float)w, 0.0f);
	glEnd();
	if (isRect)
	{
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_TEXTURE_RECTANGLE_NV);
	}
	if (texture == -1)
	{
		glEnable(GL_TEXTURE_2D);
	}
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);
	glEnable(GL_CULL_FACE);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

//rww
void GL_ResampleBloom(int w, int h, float extraBasis)
{
	float clrFrac = 0.34f + (0.04f*extraBasis);//0.34f;//0.29f;
	float texInc = 0.001f + (0.001f*extraBasis);//0.001f;//0.0015f;
	int samples = 6;//4;
	float dispRatio = (float)w/(float)h;
	float texIncH = texInc*dispRatio;
	float tex;
	float texH;
	qboolean b;
	int i;

	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);
	glDisable(GL_CULL_FACE);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, w, h, 0, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	if (!gl_lowQualityBloom.value)
	{
		g_glExt.glBindProgramARB(GL_VERTEX_PROGRAM_ARB, g_fpBloomVertH);
		glEnable(GL_VERTEX_PROGRAM_ARB);
		g_glExt.glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, g_fpBloomFragH);
		glEnable(GL_FRAGMENT_PROGRAM_ARB);

		g_glExt.glProgramEnvParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 0, clrFrac, clrFrac, clrFrac, 0.0f);
	}
	else
	{
		samples = 4;
//		texInc *= (1.0f+(g_lightExposureFactor*4.0f));
//		texIncH *= (1.0f+(g_lightExposureFactor*4.0f));
		//texIncH = texInc;
	}

	//initially clear the blur handle
	GL_BeginFBORender(g_glFBOBloomBlurHandle, -1, -1);
	glClear(GL_COLOR_BUFFER_BIT);
	GL_EndFBORender();

	tex = texInc*0.15f;
	texH = texIncH*0.15f;
	b = false;
	for (i = 0; i < samples; i++)
	{
		unsigned int nextH;
		int nextT;
		if (!gl_lowQualityBloom.value)
		{
			g_glExt.glProgramEnvParameter4fARB(GL_VERTEX_PROGRAM_ARB, 0, tex, -tex, texH, -texH);
		}

		if (!b)
		{
			nextH = g_glFBOBloomBlurHandle;
			nextT = g_glFBOBloomTex;
		}
		else
		{
			nextH = g_glFBOBloomHandle;
			nextT = g_glFBOBloomBlur;
		}

		glBindTexture(GL_TEXTURE_2D, nextT); 

		GL_BeginFBORender(nextH, -1, -1);

		if (!gl_lowQualityBloom.value)
		{
			glBegin(GL_QUADS);
				glTexCoord2f(0.0f, 1.0f); 
				glVertex2f(0.0f, 0.0f);

				glTexCoord2f(0.0f, 0.0f);
				glVertex2f(0.0f, (float)h);

				glTexCoord2f(1.0f, 0.0f);
				glVertex2f((float)w, (float)h);

				glTexCoord2f(1.0f, 1.0f);
				glVertex2f((float)w, 0.0f);
			glEnd();
		}
		else
		{
			if (b)
			{
				tex = -tex * 2;
				texH = -texH * 2;
			}
			glBegin(GL_QUADS);
				glTexCoord2f(0.0f, 1.0f); 
				glVertex2f(0.0f+(tex*w), 0.0f+(texH*h));

				glTexCoord2f(0.0f, 0.0f);
				glVertex2f(0.0f+(tex*w), (float)h+(texH*h));

				glTexCoord2f(1.0f, 0.0f);
				glVertex2f((float)w+(tex*w), (float)h+(texH*h));

				glTexCoord2f(1.0f, 1.0f);
				glVertex2f((float)w+(tex*w), 0.0f+(texH*h));
			glEnd();
			tex = -tex;
			glBegin(GL_QUADS);
				glTexCoord2f(0.0f, 1.0f); 
				glVertex2f(0.0f+(tex*w), 0.0f+(texH*h));

				glTexCoord2f(0.0f, 0.0f);
				glVertex2f(0.0f+(tex*w), (float)h+(texH*h));

				glTexCoord2f(1.0f, 0.0f);
				glVertex2f((float)w+(tex*w), (float)h+(texH*h));

				glTexCoord2f(1.0f, 1.0f);
				glVertex2f((float)w+(tex*w), 0.0f+(texH*h));
			glEnd();
			tex = -tex;
			if (b)
			{
				tex = -tex / 2;
				texH = -texH / 2;
			}
		}
		GL_EndFBORender();

		if (gl_lowQualityBloom.value)
		{
			glEnable(GL_BLEND);
			glBlendFunc(GL_ONE, GL_ONE);
			//glColor4f(0.75f, 0.75f, 0.75f, 1.0f);
			glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		}

		b = !b;
		tex += texInc;
		texH += texIncH;
	}

	if (!gl_lowQualityBloom.value)
	{
		glDisable(GL_VERTEX_PROGRAM_ARB);
		glDisable(GL_FRAGMENT_PROGRAM_ARB);
	}
	else
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDisable(GL_BLEND);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	}

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	glDepthMask(GL_TRUE);
}

//rww
void R_DrawAliasModel (entity_t *e, int lightNum);
void R_DrawViewModel (int lightNum);
void R_DrawSpriteModel (entity_t *e);
void R_DrawSunSky(qboolean occlude);
qboolean g_depthPass = false;
qboolean g_bloomin = false;

extern msurface_t *skychain;
extern msurface_t *waterchain;
int GL_DrawSceneDepth(int chaincount, qboolean doBloom, qboolean drawSky)
{
	msurface_t *s;
	texture_t *t;
	int i;

	chaincount++;
	if (chaincount == 0)
	{ //overflow
		chaincount = 1;
	}

	g_depthPass = true;
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

	//draw the world
	for (i=0 ; i<cl.worldmodel->numtextures ; i++)
	{
		t = cl.worldmodel->textures[i];
		if (!t)
			continue;
		s = t->texturechain;
		if (!s)
			continue;
		if (i == skytexturenum)
		{
			if (drawSky && !doBloom)
			{
				R_DrawSkyChain (s);
			}
		}
		else if (i == mirrortexturenum && r_mirroralpha.value != 1.0)
		{
			continue;
		}
		else
		{
			for ( ; s ; s=s->texturechain)
			{
				texture_t *t;

				if (s->chaincount == chaincount)
				{
					break;
				}
				s->chaincount = chaincount;

				if (s->flags & SURF_DRAWTURB)
				{
					EmitWaterPolys(s);
				}
				else
				{
					DrawGLPoly(NULL, 0, s->polys, NULL, NULL, false);
				}
			}
		}
	}

	//draw entities
	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		currententity = &cl_visedicts[i];

		switch (currententity->model->type)
		{
		case mod_alias:
			R_DrawAliasModel (currententity, -1);
			break;

		case mod_brush:
			R_DrawBrushModel (currententity, -1);
			break;

		case mod_rdm:
			Model_DrawRDMModel(currententity, -1);
			break;

		default:
			break;
		}
	}

	//draw view model
	R_DrawViewModel (-1);

	//surface sprites
	GL_RenderSurfSprites();

	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

	//go through ents again to draw bloomin' ents
	//this time only draw alias models and sprites
	if (doBloom)
	{
		float c[4];

		//first draw sun, and then draw sky obstructions
		if (g_hasSun && skytexturenum >= 0)
		{
			g_bloomin = true;
			R_DrawSunSky(false);
			g_bloomin = false;
			if (drawSky)
			{
				glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
				t = cl.worldmodel->textures[skytexturenum];
				if (t)
				{
					s = t->texturechain;
					if (s)
					{
						R_DrawSkyChain(s);
					}
				}
				glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
			}
		}

		c[0] = g_glClearColor[0]*0.25f;
		c[1] = g_glClearColor[1]*0.25f;
		c[2] = g_glClearColor[2]*0.25f;
		c[3] = g_glClearColor[3];
		glFogfv(GL_FOG_COLOR, c);
		glEnable(GL_FOG);

		g_bloomin = true;

		if (GL_FancyWater())
		{ //specular glow
			g_chaincount = chaincount;
			R_DrawWaterSurfaces();
			chaincount = g_chaincount;
		}

		for (i=0 ; i<cl_numvisedicts ; i++)
		{
			currententity = &cl_visedicts[i];

			switch (currententity->model->type)
			{
			case mod_alias:
				R_DrawAliasModel (currententity, -1);
				break;

			case mod_sprite:
				R_DrawSpriteModel (currententity);
				break;

			case mod_rdm: //rww
				Model_DrawRDMModel(currententity, -1);
				break;

			default:
				break;
			}
		}
		R_DrawParticles(); //glowy particles

		glDisable(GL_FOG);
		glFogfv(GL_FOG_COLOR, g_glClearColor);

		g_bloomin = false;
	}

	g_depthPass = false;

	chaincount++;
	if (chaincount == 0)
	{ //overflow
		chaincount = 1;
	}
	return chaincount;
}

//rww
extern qboolean GL_FBO_Init(void);
qboolean GL_BloomPass(void)
{
	static float extraBasis = 0.0f;
	float avgGlowFactor;

	if (!g_canUseTexRect || !g_canUseFBO)
	{ //don't have the necessary extensions
		return false;
	}

	if (g_bloomLowQuality != !!gl_lowQualityBloom.value)
	{ //change in quality, re-init FBO's
		if (!GL_FBO_Init())
		{
			g_canUseFBO = false;
			return false;
		}
	}

	//slowly decrement basis
	extraBasis -= ((float)host_frametime)*0.25f;
	if (extraBasis < 0.0f)
	{
		extraBasis = 0.0f;
	}

	if (g_lightExposureFactor > 1.0f)
	{ //cap at sane value
		g_lightExposureFactor = 1.0f;
	}
	if (extraBasis < g_lightExposureFactor)
	{
		extraBasis += (g_lightExposureFactor*0.25f);
		if (extraBasis > g_lightExposureFactor)
		{
			extraBasis = g_lightExposureFactor;
		}
	}

	avgGlowFactor = 0.2f + (0.6f*extraBasis);

	//capture the final render of the main scene to texture
	GL_CapScreenTex();

	//begin rendering scene to the bloom fbo
	GL_BeginFBORender(g_glFBOBloomHandle, -1, -1);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	//render the main scene's texture into the glow buffer, darkened in contrast to actual
	//glowing objects
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glColor4f(avgGlowFactor, avgGlowFactor, avgGlowFactor, 1.0f);
	GL_FullscreenQuad(screen_texture, window_width, window_height, true);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	//render scene depth, and then render models/sprites which glow on top of depth
	g_chaincount = GL_DrawSceneDepth(g_chaincount, true, true);

	GL_EndFBORender();

	//do the actual blurring
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	GL_ResampleBloom(g_bloomBufWidth, g_bloomBufHeight, extraBasis);

	//now render the final blur texture over the main framebuffer
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);
	GL_FullscreenQuad(g_glFBOBloomBlur, g_bloomBufWidth, g_bloomBufHeight, false);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_BLEND);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	return true;
}

/*
=============================================================

  SPRITE MODELS

=============================================================
*/

/*
================
R_GetSpriteFrame
================
*/
mspriteframe_t *R_GetSpriteFrame (entity_t *currententity)
{
	msprite_t		*psprite;
	mspritegroup_t	*pspritegroup;
	mspriteframe_t	*pspriteframe;
	int				i, numframes, frame;
	float			*pintervals, fullinterval, targettime, time;

	psprite = currententity->model->cache.data;
	frame = currententity->frame;

	if ((frame >= psprite->numframes) || (frame < 0))
	{
		Con_Printf ("R_DrawSprite: no such frame %d\n", frame);
		frame = 0;
	}

	if (psprite->frames[frame].type == SPR_SINGLE)
	{
		pspriteframe = psprite->frames[frame].frameptr;
	}
	else
	{
		pspritegroup = (mspritegroup_t *)psprite->frames[frame].frameptr;
		pintervals = pspritegroup->intervals;
		numframes = pspritegroup->numframes;
		fullinterval = pintervals[numframes-1];

		time = cl.time + currententity->syncbase;

	// when loading in Mod_LoadSpriteGroup, we guaranteed all interval values
	// are positive, so we don't have to worry about division by 0
		targettime = time - ((int)(time / fullinterval)) * fullinterval;

		for (i=0 ; i<(numframes-1) ; i++)
		{
			if (pintervals[i] > targettime)
				break;
		}

		pspriteframe = pspritegroup->frames[i];
	}

	return pspriteframe;
}


/*
=================
R_DrawSpriteModel

=================
*/
void R_DrawSpriteModel (entity_t *e)
{
	vec3_t	point;
	mspriteframe_t	*frame;
	float		*up, *right;
	vec3_t		v_forward, v_right, v_up;
	msprite_t		*psprite;

	// don't even bother culling, because it's just a single
	// polygon without a surface cache
	frame = R_GetSpriteFrame (e);
	psprite = currententity->model->cache.data;

	if (psprite->type == SPR_ORIENTED)
	{	// bullet marks on walls
		AngleVectors (currententity->angles, v_forward, v_right, v_up);
		up = v_up;
		right = v_right;
	}
	else
	{	// normal sprite
		up = vup;
		right = vright;
	}

	glColor3f (1,1,1);

	GL_DisableMultitexture();

    GL_Bind(frame->gl_texturenum);

	glEnable (GL_ALPHA_TEST);
	glBegin (GL_QUADS);

	glEnable (GL_ALPHA_TEST);
	glBegin (GL_QUADS);

	glTexCoord2f (0, 1);
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (0, 0);
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 0);
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 1);
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);
	
	glEnd ();

	glDisable (GL_ALPHA_TEST);
}

/*
=============================================================

  ALIAS MODELS

=============================================================
*/


#define NUMVERTEXNORMALS	162

float	r_avertexnormals[NUMVERTEXNORMALS][3] = {
#include "anorms.h"
};

vec3_t	shadevector;
float	shadelight, ambientlight;

// precalculated dot products for quantized angles
#define SHADEDOT_QUANT 16
float	r_avertexnormal_dots[SHADEDOT_QUANT][256] =
#include "anorm_dots.h"
;

float	*shadedots = r_avertexnormal_dots[0];

int	lastposenum;

//rww - calculates light and potentially halfangle vectors for an alias model
void GL_CalcModelLightVectors(int numVerts, modelTan_t *tangents, float *batchVerts,
							  float *lightVectors, float *vertNormals, dlight_t *dlight,
							  entity_t *e, int posenum, qboolean backwards)
{
	int i;
	float vertLightPos[3];

	if (dlight)
	{
		if (e)
		{
			VectorCopy(dlight->localPos, vertLightPos);
		}
		else
		{
			VectorCopy(dlight->origin, vertLightPos);
		}
	}
	else
	{
		if (g_hasSun)
		{
			GL_SetupCurEntityForDLight(NULL);
			VectorCopy(currententity->sunLocal, vertLightPos);
		}
		else
		{
			float a[3];

			a[PITCH] = currententity->angles[PITCH];
			a[YAW] = currententity->angles[YAW];
			a[ROLL] = 0.0f;
			AngleVectors(a, vertLightPos, 0, 0);
			VectorScale(vertLightPos, 128.0f, vertLightPos);
			vertLightPos[2] = 1024.0f;
		}
	}

	if (backwards)
	{
		for (i = 0; i < numVerts; i++)
		{
			float localLightVec[3];
			modelTan_t *tan = tangents+i;
			float *vert = batchVerts+(3*numVerts*posenum)+(i*3);
			float *lightVec = lightVectors+(i*3);

			VectorSubtract(vertLightPos, vert, localLightVec);
			lightVec[0] = DotProduct(localLightVec, tan->tangentMatrix.x1);
			lightVec[1] = DotProduct(localLightVec, tan->tangentMatrix.x2);
			lightVec[2] = -DotProduct(localLightVec, tan->tangentMatrix.x3);

			//todo - using local viewpos, set halfangles
		}
	}
	else
	{
		for (i = 0; i < numVerts; i++)
		{
			float localLightVec[3];
			modelTan_t *tan = tangents+i;
			float *vert = batchVerts+(i*3);
			float *lightVec = lightVectors+(i*3);

			VectorSubtract(vertLightPos, vert, localLightVec);
			lightVec[0] = DotProduct(localLightVec, tan->tangentMatrix.x1);
			lightVec[1] = DotProduct(localLightVec, tan->tangentMatrix.x2);
			lightVec[2] = DotProduct(localLightVec, tan->tangentMatrix.x3);

			//todo - using local viewpos, set halfangles
		}
	}
}

/*
=============
GL_DrawAliasFrame
=============
*/
extern int normalmap_default; //rww
void GL_DrawAliasFrame (aliashdr_t *paliashdr, int posenum, int normalTex, int lightNum) //rww
{
#if 1 //rww
	int normalTx = normalTex;
	dlight_t *dlight = NULL;
	if (lightNum >= 0)
	{
		dlight = &cl_dlights[lightNum];
	}

	if (normalTx <= 0 && dlight)
	{ //force a normal texture for the light path
		normalTx = normalmap_default;
	}

	if (g_canUseTexFP && gl_normalMaps.value && normalTex > 0 && !g_depthPass)
	{
		if (posenum != paliashdr->tanPoseNum)
		{ //generate tangents
			paliashdr->tanPoseNum = posenum;
			GL_CreateTangentMatrices(paliashdr->numverts, (modelVert_t *)(paliashdr->batchVerts + (3*paliashdr->numverts*posenum)), paliashdr->numtris, (modelTriFace_t *)paliashdr->tris, paliashdr->tangents,
				(modelTexCoord_t *)paliashdr->batchTC, paliashdr->vertNormals + (3*paliashdr->numverts*posenum));
		}

		GL_CalcModelLightVectors(paliashdr->numverts, paliashdr->tangents, paliashdr->batchVerts,
			paliashdr->lightVectors, paliashdr->vertNormals, dlight, currententity, posenum, true);

		if (!dlight)
		{ //bind the standard lighting program
			float f = 1.5f; //brightness hack
			g_glExt.glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, g_fpLightModelAmbientH);
			glEnable(GL_FRAGMENT_PROGRAM_ARB);

			glColor4f(shadelight*f, shadelight*f, shadelight*f, 1.0f);
		}

		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), paliashdr->batchVerts + (3*paliashdr->numverts*posenum));
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, 2 * sizeof(GL_FLOAT), paliashdr->batchTC);

		//set up the normal map texture
		g_glExt.glActiveTextureARB(GL_TEXTURE1_ARB);
		g_glExt.glClientActiveTextureARB(GL_TEXTURE1_ARB);
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, normalTx);
		//stick the light vectors in the tc for this unit, since we can just use normal tc for the nmap
		glTexCoordPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), paliashdr->lightVectors);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		glDrawElements(GL_TRIANGLES, paliashdr->numtris*3, GL_UNSIGNED_INT, paliashdr->tris);

		glDisable(GL_TEXTURE_2D); //disable 2d on texture1
		glDisableClientState(GL_TEXTURE_COORD_ARRAY); //disable tc array on texture1

		g_glExt.glActiveTextureARB(GL_TEXTURE0_ARB);
		g_glExt.glClientActiveTextureARB(GL_TEXTURE0_ARB);

		if (!dlight)
		{
			glDisable(GL_FRAGMENT_PROGRAM_ARB);
		}

		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	}
	else
	{ //crappy path, still uses arrays
		static float vertColors[3*MAXALIASVERTS];
		int i;
		float *vertNormal = paliashdr->vertNormals;
		float entVec[3];
		float a[3];

		if (!g_depthPass)
		{
			a[PITCH] = currententity->angles[PITCH];
			a[YAW] = currententity->angles[YAW];
			a[ROLL] = 0.0f;
			AngleVectors(a, entVec, 0, 0);

			for (i = 0; i < paliashdr->numverts; i++)
			{
				float f = 0.8f+(DotProduct(vertNormal, entVec)*0.2f);
				float mclr = (shadelight*f)*1.5f;
				if (mclr > 1.0f)
				{
					mclr = 1.0f;
				}
				vertColors[(i*3)+0] = mclr;
				vertColors[(i*3)+1] = mclr;
				vertColors[(i*3)+2] = mclr;
				vertNormal += 3;
			}

			glEnableClientState(GL_COLOR_ARRAY);
			glColorPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), vertColors);
		}
		else
		{
			glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		}

		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), paliashdr->batchVerts + (3*paliashdr->numverts*posenum));
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, 2 * sizeof(GL_FLOAT), paliashdr->batchTC);

		glDrawElements(GL_TRIANGLES, paliashdr->numtris*3, GL_UNSIGNED_INT, paliashdr->tris);

		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		if (!g_depthPass)
		{
			glDisableClientState(GL_COLOR_ARRAY);
		}
	}
#else
	float 	l;
	trivertx_t	*verts;
	int		*order;
	int		count;

lastposenum = posenum;

	verts = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts += posenum * paliashdr->poseverts;
	order = (int *)((byte *)paliashdr + paliashdr->commands);

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;
		if (!count)
			break;		// done
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
			glBegin (GL_TRIANGLE_STRIP);

		do
		{
			// texture coordinates come from the draw list
			glTexCoord2f (((float *)order)[0], ((float *)order)[1]);
			order += 2;

			// normals and vertexes come from the frame list
			l = shadedots[verts->lightnormalindex] * shadelight;
			glColor3f (l, l, l);
			glVertex3f (verts->v[0], verts->v[1], verts->v[2]);
			verts++;
		} while (--count);

		glEnd ();
	}
#endif
}


/*
=============
GL_DrawAliasShadow
=============
*/
extern	vec3_t			lightspot;

void GL_DrawAliasShadow (aliashdr_t *paliashdr, int posenum)
{
	trivertx_t	*verts;
	int		*order;
	vec3_t	point;
	float	height, lheight;
	int		count;

	lheight = currententity->origin[2] - lightspot[2];

	height = 0;
	verts = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts += posenum * paliashdr->poseverts;
	order = (int *)((byte *)paliashdr + paliashdr->commands);

	height = -lheight + 1.0;

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;
		if (!count)
			break;		// done
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
			glBegin (GL_TRIANGLE_STRIP);

		do
		{
			// texture coordinates come from the draw list
			// (skipped for shadows) glTexCoord2fv ((float *)order);
			order += 2;

			// normals and vertexes come from the frame list
			point[0] = verts->v[0] * paliashdr->scale[0] + paliashdr->scale_origin[0];
			point[1] = verts->v[1] * paliashdr->scale[1] + paliashdr->scale_origin[1];
			point[2] = verts->v[2] * paliashdr->scale[2] + paliashdr->scale_origin[2];

			point[0] -= shadevector[0]*(point[2]+lheight);
			point[1] -= shadevector[1]*(point[2]+lheight);
			point[2] = height;
//			height -= 0.001;
			glVertex3fv (point);

			verts++;
		} while (--count);

		glEnd ();
	}	
}



/*
=================
R_SetupAliasFrame

=================
*/
void R_SetupAliasFrame (int frame, aliashdr_t *paliashdr, int normalTex, int lightNum) //rww
{
	int				pose, numposes;
	float			interval;

	if ((frame >= paliashdr->numframes) || (frame < 0))
	{
		Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", frame);
		frame = 0;
	}

	pose = paliashdr->frames[frame].firstpose;
	numposes = paliashdr->frames[frame].numposes;

	if (numposes > 1)
	{
		interval = paliashdr->frames[frame].interval;
		pose += (int)(cl.time / interval) % numposes;
	}

	GL_DrawAliasFrame (paliashdr, pose, normalTex, lightNum); //rww
}

//rww
qboolean GL_GlowModelName(const char *str)
{
	if (!strcmp(str, "progs/flame2.mdl") ||
		!strcmp(str, "progs/flame.mdl") ||
		!strcmp(str, "progs/bolt.mdl") ||
		!strcmp(str, "progs/bolt2.mdl") ||
		!strcmp(str, "progs/bolt3.mdl"))
	{
		return true;
	}

	return false;
}

/*
=================
R_DrawAliasModel

=================
*/
void R_DrawAliasModel (entity_t *e, int lightNum)
{
	int			i;
	int			lnum;
	vec3_t		dist;
	float		add;
	model_t		*clmodel;
	aliashdr_t	*paliashdr;
	float		an;
	int			anim;
	int			nrmlTex;
	qboolean isTorch = false;
	qboolean isBloomin = false;

	clmodel = currententity->model;

	if (e->cullFrame != r_framecount)
	{ //rww
		vec3_t		mins, maxs;
		VectorAdd (currententity->origin, clmodel->mins, mins);
		VectorAdd (currententity->origin, clmodel->maxs, maxs);

		/*
		if (1)
		{
			float clr[3] = {1.0f, 0.0f, 0.0f};
			GL_DebugDrawLineMinsMaxs(mins, maxs, clr);
		}
		*/

		e->culled = R_CullBox(mins, maxs);
		e->cullFrame = r_framecount;
	}

	if (e->culled)
	{ //rww - since these things go through passes for so many things, only do a cull check once per frame
		return;
	}

	if (GL_GlowModelName(clmodel->name))
	{
		isTorch = true;
		if (!strcmp(clmodel->name, "progs/flame.mdl"))
		{ //hack, don't glow torch with handle. should use different masked texture for that model.
			isBloomin = false;
		}
		else
		{
			isBloomin = true;
		}
	}

	if (g_depthPass && g_bloomin != isBloomin)
	{ //neglect to draw for now
		return;
	}

	VectorCopy (currententity->origin, r_entorigin);
	VectorSubtract (r_origin, r_entorigin, modelorg);

	//
	// get lighting information
	//
	if (lightNum < 0 && !g_depthPass)
	{ //rww
		ambientlight = shadelight = R_LightPoint (currententity->origin);

		// allways give the gun some light
		if (e == &cl.viewent && ambientlight < 24)
			ambientlight = shadelight = 24;

		if (!g_canUseTexFP || !gl_normalMaps.value)
		{ //rww - don't factor dlights in here when using good lighting path
			for (lnum=0 ; lnum<g_numLiveLights ; lnum++)
			{
				VectorSubtract (currententity->origin,
								g_liveLights[lnum]->origin,
								dist);
				add = g_liveLights[lnum]->radius - Length(dist);

				if (add > 0) {
					ambientlight += add;
					//ZOID models should be affected by dlights as well
					shadelight += add;
				}
			}
		}

		// clamp lighting so it doesn't overbright as much
		if (ambientlight > 128)
			ambientlight = 128;
		if (ambientlight + shadelight > 192)
			shadelight = 192 - ambientlight;

		// ZOID: never allow players to go totally black
		if (!strcmp(clmodel->name, "progs/player.mdl")) {
			if (ambientlight < 8)
				ambientlight = shadelight = 8;

		}
		else if (isTorch)
		{
			// HACK HACK HACK -- no fullbright colors, so make torches full light
			ambientlight = shadelight = 256;
		}

		shadedots = r_avertexnormal_dots[((int)(e->angles[1] * (SHADEDOT_QUANT / 360.0))) & (SHADEDOT_QUANT - 1)];
		shadelight = shadelight / 200.0;
		
		an = e->angles[1]/180*M_PI;
		shadevector[0] = cos(-an);
		shadevector[1] = sin(-an);
		shadevector[2] = 1;
		VectorNormalize (shadevector);
	}

	//
	// locate the proper data
	//
	paliashdr = (aliashdr_t *)Mod_Extradata (currententity->model);

	c_alias_polys += paliashdr->numtris;

	//
	// draw all the triangles
	//

	GL_DisableMultitexture();

    glPushMatrix ();
	R_RotateForEntity (e);

	/*if (!strcmp (clmodel->name, "progs/eyes.mdl") ) {
		glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2] - (22 + 8));
	// double size of eyes, since they are really hard to see in gl
		glScalef (paliashdr->scale[0]*2, paliashdr->scale[1]*2, paliashdr->scale[2]*2);
	} else*/ { //rww - don't care about eye size hack
		glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2]);
		glScalef (paliashdr->scale[0], paliashdr->scale[1], paliashdr->scale[2]);
	}

	anim = (int)(cl.time*10) & 3;
#ifdef _DEBUG //rww
	if (!paliashdr->gl_texturenum[currententity->skinnum][anim]) {
		Con_Printf("Bad texture index for model %s\n", clmodel->name);
	}
#endif
    GL_Bind(paliashdr->gl_texturenum[currententity->skinnum][anim]);
	nrmlTex = paliashdr->gl_normaltexturenum[currententity->skinnum][anim];

	// we can't dynamically colormap textures, so they are cached
	// seperately for the players.  Heads are just uncolored.
	if (currententity->scoreboard && !gl_nocolors.value)
	{
		i = currententity->scoreboard - cl.players;
		if (!currententity->scoreboard->skin) {
			Skin_Find(currententity->scoreboard);
			R_TranslatePlayerSkin(i);
		}
		if (i >= 0 && i<MAX_CLIENTS)
		{
		    GL_Bind(playertextures + (i*2)); //rww - *2
			nrmlTex = playertextures + (i*2) + 1; //rww
		}
	}

	if (gl_smoothmodels.value)
		glShadeModel (GL_SMOOTH);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	if (gl_affinemodels.value)
		glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

	R_SetupAliasFrame (currententity->frame, paliashdr, nrmlTex, lightNum); //rww

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glShadeModel (GL_FLAT);
	if (gl_affinemodels.value)
		glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	glPopMatrix ();

	if (lightNum < 0 && r_shadows.value)
	{ //rww - don't do shadows on light passes
		glPushMatrix ();
		R_RotateForEntity (e);
		glDisable (GL_TEXTURE_2D);
		glEnable (GL_BLEND);
		glColor4f (0,0,0,0.5);
		GL_DrawAliasShadow (paliashdr, lastposenum);
		glEnable (GL_TEXTURE_2D);
		glDisable (GL_BLEND);
		glColor4f (1,1,1,1);
		glPopMatrix ();
	}

}

//==================================================================================

const int r_ambStenInRef = 0;
const int r_ambStenRef = 0;
const int r_ambStenMask = 0xFFFFFFFF;

//rww
void R_AmbientShadowBegin(void)
{
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_STENCIL_TEST);
#if 0
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);
	glColor4f(0.05f, 0.05f, 0.05f, 1.0f);
#else
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
#endif
	glDepthMask(GL_FALSE);
	glStencilFunc(GL_ALWAYS, r_ambStenRef, r_ambStenMask);
	glDepthFunc(GL_LESS);
	glEnable(GL_CULL_FACE);
}

//rww
void GL_FullscreenQuad(int texture, int w, int h, qboolean isRect);
void R_AmbientShadowEnd(float fade)
{
	glCullFace(GL_FRONT);
	glDepthFunc(GL_LEQUAL);
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glStencilFunc(GL_NOTEQUAL, r_ambStenInRef, r_ambStenMask);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

#if 0
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_BLEND);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
#endif

	//fill shadowed areas with a big stupid fullscreen quad
	glEnable(GL_BLEND);
	glColor4f(0.0f, 0.0f, 0.0f, fade);
	GL_FullscreenQuad(-1, window_width, window_height, false); //reenables texture 2d and depth mask
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glDisable(GL_BLEND);

	glDisable(GL_STENCIL_TEST);
}

/*
//rww
void R_AmbientShadowNegate(void)
{
	glEnable(GL_STENCIL_TEST);
	glCullFace(GL_FRONT);
	glDepthFunc(GL_LEQUAL);
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	glStencilFunc(GL_NOTEQUAL, r_ambStenInRef, r_ambStenMask);
	glStencilOp(GL_KEEP, GL_ZERO, GL_ZERO);

	glDisable(GL_TEXTURE_2D);
	glDepthMask(GL_FALSE);
	//glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
	GL_FullscreenQuad(-1, window_width, window_height, false); //reenables texture 2d and depth mask
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glDisable(GL_STENCIL_TEST);
}

//rww
typedef struct ambShadowFade_s
{
	entity_t		ent;
	double			time;
	float			val;
	int				visframe;
	qboolean		hit;
} ambShadowFade_t;
#define MAX_SHADOW_FADES (MAX_EDICTS+MAX_STATIC_ENTITIES)
static ambShadowFade_t r_ambShadowFades[MAX_SHADOW_FADES];

//rww
void R_ResetShadowFades(void)
{
	memset(r_ambShadowFades, 0, sizeof(r_ambShadowFades));
}

//rww
extern float g_lodHackDist;
qboolean R_ShouldShadowFade(entity_t *e)
{
	ambShadowFade_t *f;
	if (e->gameentnum < 0 || e->gameentnum >= MAX_SHADOW_FADES)
	{
		return false;
	}
	f = &r_ambShadowFades[e->gameentnum];
	if (!f->hit)
	{
		f->hit = true;
	}

	if (e->model && e->model->type == mod_rdm)
	{ //check for lodhack for trees and such
		modelFile_t *model = Model_CoreFromQuakeEnt(e);
		if (model && model->lodHack)
		{
			if (model->lodDist > g_lodHackDist)
			{
				e->sunCulled = true;
			}
		}
	}

	f->ent = *e;

	f->visframe = r_framecount;

	if (f->val < 1.0f)
	{
		return true;
	}
	
	return false;
}
*/

//rww
void GL_DrawAliasVolume(entity_t *e, int lightNum);
void R_DrawCurrentEntVolume(void)
{
	switch (currententity->model->type)
	{
	case mod_alias:
		GL_SetupCurEntityForDLight(NULL);
		GL_DrawAliasVolume(currententity, -1);
		break;
	case mod_rdm:
		GL_SetupCurEntityForDLight(NULL);
		GL_DrawRDMVolume(currententity, -1);
		break;
	case mod_brush:
		//screw it, too many issues with these
		/*
		GL_SetupCurEntityForDLight(NULL);
		GL_DrawBrushModelVolume(currententity, -1);
		*/
		break;

	default:
		break;
	}
}

//rww - ambient sun volumes for alias models
void R_DrawEntitySunVolumes(void)
{
	int i;
	float d[3], l;

	if (!g_hasSun || !gl_ambientShadows.value)
	{ //no sun or disabled ambient shadows
		return;
	}

	R_AmbientShadowBegin();

	if (g_twoSidedStencil && gl_twoSidedStencil.value)
	{
		glEnable(GL_STENCIL_TEST_TWO_SIDE_EXT);

		g_glExt.glActiveStencilFaceEXT(GL_BACK);
		glStencilOp(GL_KEEP, GL_DECR_WRAP_EXT, GL_KEEP);

		g_glExt.glActiveStencilFaceEXT(GL_FRONT);
		glStencilOp(GL_KEEP, GL_INCR_WRAP_EXT, GL_KEEP);

		glDisable(GL_CULL_FACE);

		for (i=0 ; i<cl_numvisedicts ; i++)
		{
			currententity = &cl_visedicts[i];

			if (currententity->model->type == mod_brush)
			{
				float pos[3];
				pos[0] = currententity->model->mins[0] + (currententity->model->maxs[0]-currententity->model->mins[0])/2.0f;
				pos[1] = currententity->model->mins[1] + (currententity->model->maxs[1]-currententity->model->mins[1])/2.0f;
				pos[2] = currententity->model->mins[2] + (currententity->model->maxs[2]-currententity->model->mins[2])/2.0f;
				VectorSubtract(r_refdef.vieworg, pos, d);
			}
			else
			{
				VectorSubtract(r_refdef.vieworg, currententity->origin, d);
			}
			l = Length(d);
			if (l+currententity->model->radius >= g_beginLodCutoff)
			{ //don't draw if beyond certain distance
				currententity->sunCulled = true;
				//R_ShouldShadowFade(currententity);
				continue;
			}

			currententity->sunCulled = false;

			/*
			if (R_ShouldShadowFade(currententity))
			{
				continue;
			}
			*/

			R_DrawCurrentEntVolume();
		}

		glDisable(GL_STENCIL_TEST_TWO_SIDE_EXT);
		glEnable(GL_CULL_FACE);
	}
	else
	{
		//first pass
		glCullFace(GL_FRONT);
		glStencilOp(GL_KEEP, GL_INCR, GL_KEEP);
		for (i=0 ; i<cl_numvisedicts ; i++)
		{
			currententity = &cl_visedicts[i];

			if (currententity->model->type == mod_brush)
			{
				float pos[3];
				pos[0] = currententity->model->mins[0] + (currententity->model->maxs[0]-currententity->model->mins[0])/2.0f;
				pos[1] = currententity->model->mins[1] + (currententity->model->maxs[1]-currententity->model->mins[1])/2.0f;
				pos[2] = currententity->model->mins[2] + (currententity->model->maxs[2]-currententity->model->mins[2])/2.0f;
				VectorSubtract(r_refdef.vieworg, pos, d);
			}
			else
			{
				VectorSubtract(r_refdef.vieworg, currententity->origin, d);
			}
			l = Length(d);
			if (l+currententity->model->radius >= g_beginLodCutoff)
			{ //don't draw if beyond certain distance
				currententity->sunCulled = true;
				//R_ShouldShadowFade(currententity);
				continue;
			}

			currententity->sunCulled = false;

			/*
			if (R_ShouldShadowFade(currententity))
			{
				continue;
			}
			*/

			R_DrawCurrentEntVolume();
		}

		//second pass
		glCullFace(GL_BACK);
		glStencilOp(GL_KEEP, GL_DECR, GL_KEEP);
		for (i=0 ; i<cl_numvisedicts ; i++)
		{
			currententity = &cl_visedicts[i];

			if (currententity->sunCulled)
			{ //culled on the first pass
				continue;
			}

			/*
			if (R_ShouldShadowFade(currententity))
			{
				continue;
			}
			*/

			R_DrawCurrentEntVolume();
		}
	}

	R_AmbientShadowEnd(0.4f);

	//fixme, probably need to useoffscreen buffer to test fragments in order to prevent
	//doubling over on two blending objects
	/*
	//check for separate-pass fading volumes
	for (i = 0; i < MAX_SHADOW_FADES; i++)
	{
		ambShadowFade_t *f = &r_ambShadowFades[i];
		float incStep = 0.05f;
		float goal;

		if (!f->hit)
		{
			continue;
		}

		if (f->visframe != r_framecount || f->ent.sunCulled)
		{
			goal = 0.0f;
		}
		else
		{
			goal = 1.0f;
		}

		if (f->val == goal)
		{
			continue;
		}

		if (f->val < goal)
		{
			f->val += incStep;
			if (f->val > goal)
			{
				f->val = goal;
			}
		}
		else if (f->val > goal)
		{
			f->val -= incStep;
			if (f->val < goal)
			{
				f->val = goal;
			}
		}

		currententity = &f->ent;

		R_AmbientShadowNegate();

		R_AmbientShadowBegin();
		if (g_twoSidedStencil && gl_twoSidedStencil.value)
		{
			glEnable(GL_STENCIL_TEST_TWO_SIDE_EXT);

			g_glExt.glActiveStencilFaceEXT(GL_BACK);
			glStencilOp(GL_KEEP, GL_DECR_WRAP_EXT, GL_KEEP);

			g_glExt.glActiveStencilFaceEXT(GL_FRONT);
			glStencilOp(GL_KEEP, GL_INCR_WRAP_EXT, GL_KEEP);

			glDisable(GL_CULL_FACE);

			R_DrawCurrentEntVolume();

			glDisable(GL_STENCIL_TEST_TWO_SIDE_EXT);
			glEnable(GL_CULL_FACE);
		}
		else
		{
			//first pass
			glCullFace(GL_FRONT);
			glStencilOp(GL_KEEP, GL_INCR, GL_KEEP);
			R_DrawCurrentEntVolume();

			//second pass
			glCullFace(GL_BACK);
			glStencilOp(GL_KEEP, GL_DECR, GL_KEEP);
			R_DrawCurrentEntVolume();
		}

		R_AmbientShadowEnd(0.4f*f->val);
	}
	*/

	glClear(GL_STENCIL_BUFFER_BIT);
}

//rww
typedef struct treeLeaf_s
{
	float		pos[3];
	float		ang[3];
	float		vel[3];
	double		time;
	float		rnd;
	float		rnd2;
	qboolean	hitGround;
	float		lifeTime;
	qboolean	alive;
} treeLeaf_t;
#define MAX_TREE_LEAFS 256
treeLeaf_t g_treeLeafs[MAX_TREE_LEAFS];
#define TREE_LEAF_GRAV	0.1f

//rww
treeLeaf_t *R_OldestTreeLeaf(void)
{
	treeLeaf_t *l = &g_treeLeafs[0];
	int i;
	for (i = 1; i < MAX_TREE_LEAFS; i++)
	{
		treeLeaf_t *n = &g_treeLeafs[i];
		if (n->time < l->time)
		{
			l = n;
		}
	}

	return l;
}

//rww
pmtrace_t PM_GeneralTrace(vec3_t start, vec3_t end, qboolean thirdHull);
void R_SimulateLeaf(treeLeaf_t *l, float timeFrac)
{
	float oldPos[3];
	float leafSpeed = 1.0f;
	pmtrace_t trace;

	if (!l->hitGround)
	{ //fake a leafy motion
		l->vel[0] += sinf(realtime*(2.5f+l->rnd))*l->rnd*(0.2f);
		l->vel[1] += cosf(realtime*(2.5f+l->rnd))*l->rnd*(0.2f);
	}

	VectorCopy(l->pos, oldPos);
	l->pos[0] += l->vel[0]*(leafSpeed*timeFrac);
	l->pos[1] += l->vel[1]*(leafSpeed*timeFrac);
	l->pos[2] += l->vel[2]*(leafSpeed*timeFrac);
	l->vel[0] *= (1.0f-(timeFrac*0.1f));
	l->vel[1] *= (1.0f-(timeFrac*0.1f));
	l->vel[2] *= (1.0f-(timeFrac*0.1f));
	if (!l->hitGround)
	{
		l->vel[2] -= (TREE_LEAF_GRAV+(l->rnd2*0.1f))*timeFrac;
	}
	trace = PM_GeneralTrace(oldPos, l->pos, true);
	if (trace.startsolid)
	{ //remove
		l->time = 0;
		return;
	}
	if (trace.fraction != 1.0f)
	{
		float dp = DotProduct(l->vel, trace.plane.normal);
		float up[3] = {0.0f, 0.0f, 1.0f};
		float u = DotProduct(up, trace.plane.normal);
		l->vel[0] -= dp*trace.plane.normal[0];
		l->vel[1] -= dp*trace.plane.normal[1];
		l->vel[2] -= dp*trace.plane.normal[2];
		if (u > 0.8f)
		{
			l->hitGround = true;
		}
		VectorCopy(trace.endpos, l->pos);
		l->pos[0] += trace.plane.normal[0];
		l->pos[1] += trace.plane.normal[1];
		l->pos[2] += trace.plane.normal[2];
	}
}

//rww
extern int hr_leaf;
float V_InterpolateAngle(float old, float cur, float val, float thresh); //borrowing
void R_DrawTreeLeafs(void)
{
	int i;
	float f = host_frametime*30.0f;
	float baseShade = 0.3f;
	float foldShade = 0.15f;
	float w = 8.0f;
	float h = 8.0f;
	float foldFactor = 4.0f;
	float fadeTime = 2.0f;
	float fade = 1.0f;
	for (i = 0; i < MAX_TREE_LEAFS; i++)
	{
		treeLeaf_t *l = &g_treeLeafs[i];
		float ang[3];
		if (!l->alive)
		{
			continue;
		}
		if ((l->time+l->lifeTime) < realtime)
		{
			l->alive = false;
			continue;
		}

		R_SimulateLeaf(l, f);

		if ((l->time+l->lifeTime)-fadeTime < realtime)
		{
			float d = realtime - ((l->time+l->lifeTime)-fadeTime);
			fade = 1.0f-(d/fadeTime);
			if (fade > 1.0f)
			{
				fade = 1.0f;
			}
			else if (fade < 0.0f)
			{
				fade = 0.0f;
			}
		}
		else
		{
			fade = 1.0f;
		}

		vectoangles(l->vel, ang);
		l->ang[0] = V_InterpolateAngle(l->ang[0], ang[0], 0.2f, 360.0f);
		l->ang[1] = V_InterpolateAngle(l->ang[1], ang[1], 0.2f, 360.0f);
		l->ang[2] = V_InterpolateAngle(l->ang[2], ang[2], 0.2f, 360.0f);

		if (fade < 1.0f)
		{
			glEnable(GL_BLEND);
		}
		else
		{
			glEnable(GL_ALPHA_TEST);
		}
		glDisable(GL_CULL_FACE);
		glPushMatrix();
		glTranslatef(l->pos[0],  l->pos[1],  l->pos[2]);
		glRotatef(l->ang[1], 0, 0, 1);
		glRotatef(-l->ang[0], 0, 1, 0);
		glRotatef(l->ang[2], 1, 0, 0);

		glShadeModel(GL_SMOOTH);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		GL_Bind(hr_leaf);
		//draw a v-cup shape
		glBegin(GL_QUADS);
			glColor4f(baseShade*fade, baseShade*fade, baseShade*fade, fade);
			glTexCoord2f(0.0f, 1.0f); 
			glVertex3f(-w*0.5f, 0.0f, foldFactor);
			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-w*0.5f, h, foldFactor);
			glColor4f(foldShade*fade, foldShade*fade, foldShade*fade, fade);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(0.0f, h, 0.0f);
			glTexCoord2f(0.5f, 1.0f);
			glVertex3f(0.0f, 0.0f, 0.0f);

			glTexCoord2f(0.5f, 1.0f); 
			glVertex3f(0.0f, 0.0f, 0.0f);
			glTexCoord2f(0.5f, 0.0f);
			glVertex3f(0.0f, h, 0.0f);
			glColor4f(baseShade*fade, baseShade*fade, baseShade*fade, fade);
			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(w*0.5f, h, foldFactor);
			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(w*0.5f, 0.0f, foldFactor);
		glEnd();
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glShadeModel(GL_FLAT);

		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		glPopMatrix();
		glEnable(GL_CULL_FACE);
		if (fade < 1.0f)
		{
			glDisable(GL_BLEND);
		}
		else
		{
			glDisable(GL_ALPHA_TEST);
		}
	}
}

//rww
void R_FracBiasVector(float *a, float *b, float frac, float *out)
{
	float t[3];
	VectorScale(a, 1.0f-frac, out);
	VectorScale(b, frac, t);
	VectorAdd(out, t, out);
}

//rww
void R_RandomPointInTri(float *v1, float *v2, float *v3, float *out)
{
	float edgeVec1[3], edgeVec2[3], t[3];
	float r1 = (rand()&0x7fff) / ((float)0x7fff);
	float r2 = (rand()&0x7fff) / ((float)0x7fff);
	float f;

	VectorSubtract(v2, v1, edgeVec1);
	VectorSubtract(v3, v1, edgeVec2);

	//get a vector between one of the edges based on the random frac, then use a point along
	//that line as the random position
	R_FracBiasVector(edgeVec1, edgeVec2, r1, t);
	f = VectorNormalize(t);
	VectorMA(v1, f*r2, t, out);
}

//rww
void R_AddTreeLeaf(float *pos)
{
	treeLeaf_t *l = R_OldestTreeLeaf();
	if (!l)
	{
		return;
	}

	memset(l, 0, sizeof(treeLeaf_t));
	l->alive = true;
	l->time = realtime;
	l->lifeTime = 8.0 + ((float)(rand()%8192)/8192.0f)*3.0f;
	l->rnd = (float)(rand()%8192)/8192.0f;
	l->rnd2 = (float)(rand()%8192)/8192.0f;
	VectorCopy(pos, l->pos);
}

//rww - for tree (of the forest variety) stuff
extern modelMatrix_t g_identityMatrix;
extern float g_lodHackDist;
void R_CheckRDMHacks(entity_t *ent)
{
	modelFile_t *model = Model_CoreFromQuakeEnt(ent);
	modelMeshObject_t *mesh;
	modelBone_t *treeBones[3];
	float framefactor = 1.0f;
	float treeBoneTransPos[3][3];
	float transMat[16], invTransMat[16];
	float windIdle[3];
	int i;
	if (!model)
	{
		return;
	}

	if (model->lodHack)
	{
		float d[3];
		VectorSubtract(r_refdef.vieworg, ent->origin, d);
		model->lodDist = Length(d);
		if (model->lodDist > g_lodHackDist || !gl_qualityFoliage.value)
		{
			return;
		}
	}

	if (!model->treeHack)
	{
		return;
	}
	mesh = model->firstObject;
	if (!mesh)
	{
		return;
	}

	windIdle[0] = sinf(cl.time)*0.5f;
	windIdle[1] = cosf(cl.time)*0.5f;
	windIdle[2] = sinf(cl.time)*0.5f;

	//get the matrix to process bone positions
	glPushMatrix();
	glLoadIdentity();
	R_RotateForEntity(ent);
	glGetFloatv(GL_MODELVIEW_MATRIX, transMat);
	glPopMatrix();
	MatrixInverse4x4(transMat, invTransMat);

	for (i = 0; i < 3; i++)
	{
		float v[3];
		char boneName[256];
		sprintf(boneName, "Bone%02i", i+2);
		treeBones[i] = Model_FindObjectBoneByName(mesh, boneName);
		if (!treeBones[i])
		{
			return;
		}
		VectorScale(treeBones[i]->eData.finalMat.o, model->baseScale, v);
		TransformPointByMatrix4x4(transMat, v, treeBoneTransPos[i]);
	}

	for (i = 0; i < 3; i++)
	{
		float locOff[3];
		int j, lnum;
		modelMatrix_t ovMat;
		modelMeshObject_t *nMesh;

		model->treeBoneVel[i][0] += windIdle[0]*framefactor;
		model->treeBoneVel[i][1] += windIdle[1]*framefactor;
		model->treeBoneVel[i][2] += windIdle[2]*framefactor;

		model->treeBoneEff[i][0] += model->treeBoneVel[i][0]*(0.25f*framefactor);
		model->treeBoneEff[i][1] += model->treeBoneVel[i][1]*(0.25f*framefactor);
		model->treeBoneEff[i][2] += model->treeBoneVel[i][2]*(0.25f*framefactor);
		model->treeBoneVel[i][0] *= (0.9f);
		model->treeBoneVel[i][1] *= (0.9f);
		model->treeBoneVel[i][2] *= (0.9f);

		model->treeBoneVel[i][0] -= model->treeBoneEff[i][0]*(0.2f*framefactor);
		model->treeBoneVel[i][1] -= model->treeBoneEff[i][1]*(0.2f*framefactor);
		model->treeBoneVel[i][2] -= model->treeBoneEff[i][2]*(0.2f*framefactor);

		//impose some sane limits on wind effect
		for (j = 0; j < 3; j++)
		{
			if (model->treeBoneEff[i][j] > 60.0f)
			{
				model->treeBoneEff[i][j] = 60.0f;
				model->treeBoneVel[i][j] = 0.0f;
			}
			else if (model->treeBoneEff[i][j] < -60.0f)
			{
				model->treeBoneEff[i][j] = -60.0f;
				model->treeBoneVel[i][j] = 0.0f;
			}
		}

		//add to vel based on positions of spriteEff lights relative to treeBoneTransPos[i]
		for (lnum=0 ; lnum<g_numLiveLights ; lnum++)
		{
			vec3_t dist;
			float add;
			if (!g_liveLights[lnum]->spriteEff)
			{
				continue;
			}

			VectorSubtract (treeBoneTransPos[i],
							g_liveLights[lnum]->origin,
							dist);
			add = Length(dist)/g_liveLights[lnum]->radius;

			if (add < 1.0f)
			{
				if (add < 0.0f)
				{
					add = 0.0f;
				}
				add = 1.0f-add;

				if (model->treeSoundTime < realtime)
				{
					float lPos[3];
					char *snd;
					int r = rand()%9;
					int n = (int)(add*32.0f);
					int k = 0;
					if (r < 3)
					{
						snd = "misc/leafs.wav";
					}
					else if (r < 6)
					{
						snd = "misc/leafs2.wav";
					}
					else
					{
						snd = "misc/leafs3.wav";
					}

					S_StartSound(ent->gameentnum, model->treeSoundChan, S_PrecacheSound(snd), treeBoneTransPos[i], add, 1.5f);
					model->treeSoundTime = realtime+0.2 + ((float)(rand()%8192)/8192.0f)*0.2f;
					model->treeSoundChan++;
					if (model->treeSoundChan >= 8)
					{
						model->treeSoundChan = 0;
					}

					for (k = 0; k < n; k++)
					{
						R_RandomPointInTri(treeBoneTransPos[0], treeBoneTransPos[1], treeBoneTransPos[2], lPos);
						R_AddTreeLeaf(lPos);
					}
				}

				model->treeBoneVel[i][0] += dist[0]*add*0.05f;
				model->treeBoneVel[i][1] += dist[1]*add*0.05f;
				model->treeBoneVel[i][2] += dist[2]*add*0.05f;
			}
		}

		locOff[0] = invTransMat[0]*model->treeBoneEff[i][0] + invTransMat[4+0]*model->treeBoneEff[i][1] + invTransMat[8+0]*model->treeBoneEff[i][2];
		locOff[1] = invTransMat[1]*model->treeBoneEff[i][0] + invTransMat[4+1]*model->treeBoneEff[i][1] + invTransMat[8+1]*model->treeBoneEff[i][2];
		locOff[2] = invTransMat[2]*model->treeBoneEff[i][0] + invTransMat[4+2]*model->treeBoneEff[i][1] + invTransMat[8+2]*model->treeBoneEff[i][2];
		VectorScale(locOff, 1.0f/model->baseScale, locOff);

		ovMat = g_identityMatrix;
		VectorCopy(locOff, ovMat.o);
		for (nMesh = mesh; nMesh; nMesh = nMesh->next)
		{
			Model_SetOverrideMatrix(nMesh, treeBones[i], &ovMat);
		}

		//for debugging
		/*
		glPointSize(15.0f);
		glDisable(GL_TEXTURE_2D);
		glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
		glBegin(GL_POINTS);
			glVertex3fv(treeBoneTransPos[i]);
		glEnd();
		glEnable(GL_TEXTURE_2D);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glPointSize(1.0f);
		*/
	}
}

/*
=============
R_DrawEntitiesOnList
=============
*/
void R_DrawEntitiesOnList (void)
{
	int		i;

	if (!r_drawentities.value)
		return;

	// draw sprites seperately, because of alpha blending
	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		currententity = &cl_visedicts[i];

		switch (currententity->model->type)
		{
		case mod_alias:
			R_DrawAliasModel (currententity, -1);
			break;

		case mod_brush:
			R_DrawBrushModel (currententity, -1);
			break;

		case mod_rdm: //rww
			R_CheckRDMHacks(currententity);
			Model_DrawRDMModel(currententity, -1);
			break;

		default:
			break;
		}
	}

	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		currententity = &cl_visedicts[i];

		switch (currententity->model->type)
		{
		case mod_sprite:
			R_DrawSpriteModel (currententity);
			break;

		default :
			break;
		}
	}
}

/*
=============
R_DrawViewModel
=============
*/
void R_DrawViewModel (int lightNum) //rww
{
	float		ambient[4], diffuse[4];
	int			j;
	int			lnum;
	vec3_t		dist;
	float		add;
	dlight_t	*dl;
	int			ambientlight, shadelight;
	dlight_t	*dlight = NULL; //rww

	if (!r_drawviewmodel.value || !Cam_DrawViewModel())
		return;

	if (envmap)
		return;

	if (!r_drawentities.value)
		return;

	if (cl.stats[STAT_ITEMS] & IT_INVISIBILITY)
		return;

	if (cl.stats[STAT_HEALTH] <= 0)
		return;

	currententity = &cl.viewent;
	if (!currententity->model)
		return;

	if (lightNum >= 0)
	{ //rww
		dlight = &cl_dlights[lightNum];
	}

	if (dlight)
	{ //rww
		if (!GL_AliasModelInLight(currententity, dlight))
		{
			return;
		}
		else
		{
			GL_SetupCurEntityForDLight(dlight);
		}
	}

	if (!g_depthPass)
	{
		j = R_LightPoint (currententity->origin);

		if (j < 24)
			j = 24;		// allways give some light on gun
		ambientlight = j;
		shadelight = j;
	}

// add dynamic lights		
	if (!g_canUseTexFP || !gl_normalMaps.value)
	{ //rww - not in good lighting mode
		for (lnum=0 ; lnum<g_numLiveLights ; lnum++)
		{
			dl = g_liveLights[lnum];
			if (!dl->radius)
				continue;
			if (!dl->radius)
				continue;

			VectorSubtract (currententity->origin, dl->origin, dist);
			add = dl->radius - Length(dist);
			if (add > 0)
				ambientlight += add;
		}
	}

	if (!g_depthPass)
	{
		ambient[0] = ambient[1] = ambient[2] = ambient[3] = (float)ambientlight / 128;
		diffuse[0] = diffuse[1] = diffuse[2] = diffuse[3] = (float)shadelight / 128;
	}

	// hack the depth range to prevent view model from poking into walls
	glDepthRange (gldepthmin, gldepthmin + 0.3*(gldepthmax-gldepthmin));
	R_DrawAliasModel (currententity, lightNum);
	glDepthRange (gldepthmin, gldepthmax);
}


/*
============
R_PolyBlend
============
*/
void R_PolyBlend (void)
{
	if (!gl_polyblend.value)
		return;
	if (!v_blend[3])
		return;

//Con_Printf("R_PolyBlend(): %4.2f %4.2f %4.2f %4.2f\n",v_blend[0], v_blend[1],	v_blend[2],	v_blend[3]);

 	GL_DisableMultitexture();

	glDisable (GL_ALPHA_TEST);
	glEnable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_TEXTURE_2D);

    glLoadIdentity ();

    glRotatef (-90,  1, 0, 0);	    // put Z going up
    glRotatef (90,  0, 0, 1);	    // put Z going up

	glColor4fv (v_blend);

	glBegin (GL_QUADS);

	glVertex3f (10, 100, 100);
	glVertex3f (10, -100, 100);
	glVertex3f (10, -100, -100);
	glVertex3f (10, 100, -100);
	glEnd ();

	glDisable (GL_BLEND);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_ALPHA_TEST);
}


int SignbitsForPlane (mplane_t *out)
{
	int	bits, j;

	// for fast box on planeside test

	bits = 0;
	for (j=0 ; j<3 ; j++)
	{
		if (out->normal[j] < 0)
			bits |= 1<<j;
	}
	return bits;
}


extern cvar_t gl_forceWideAspect;
void R_SetFrustum (void)
{
	int		i;

	if (r_refdef.fov_x == 90) 
	{
		// front side is visible

		VectorAdd (vpn, vright, frustum[0].normal);
		VectorSubtract (vpn, vright, frustum[1].normal);

		VectorAdd (vpn, vup, frustum[2].normal);
		VectorSubtract (vpn, vup, frustum[3].normal);
	}
	else
	{

		// rotate VPN right by FOV_X/2 degrees
		RotatePointAroundVector( frustum[0].normal, vup, vpn, -(90-r_refdef.fov_x / 2 ) );
		// rotate VPN left by FOV_X/2 degrees
		RotatePointAroundVector( frustum[1].normal, vup, vpn, 90-r_refdef.fov_x / 2 );
		// rotate VPN up by FOV_X/2 degrees
		RotatePointAroundVector( frustum[2].normal, vright, vpn, 90-r_refdef.fov_y / 2 );
		// rotate VPN down by FOV_X/2 degrees
		RotatePointAroundVector( frustum[3].normal, vright, vpn, -( 90 - r_refdef.fov_y / 2 ) );
	}

	for (i=0 ; i<4 ; i++)
	{
		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct (r_origin, frustum[i].normal);
		frustum[i].signbits = SignbitsForPlane (&frustum[i]);
	}
}



/*
===============
R_SetupFrame
===============
*/
void R_SetupFrame (void)
{
// don't allow cheats in multiplayer
	r_fullbright.value = 0;
	r_lightmap.value = 0;
	if (!atoi(Info_ValueForKey(cl.serverinfo, "watervis")))
	{
		//r_wateralpha.value = 1;
		//rww - set it back right
		Cvar_Set("r_wateralpha", "1");
	}
	else if (GL_FancyWater())
	{ //rww
		if (r_wateralpha.value != 0.6f)
		{
			Cvar_Set("r_wateralpha", "0.6");
		}
	}

	R_AnimateLight ();

	r_framecount++;

	g_numSurfSpriteDraws = 0; //rww

// build the transformation matrix for the given view angles
	VectorCopy (r_refdef.vieworg, r_origin);

	AngleVectors (r_refdef.viewangles, vpn, vright, vup);

// current viewleaf
	r_oldviewleaf = r_viewleaf;
	r_viewleaf = Mod_PointInLeaf (r_origin, cl.worldmodel);

	V_SetContentsColor (r_viewleaf->contents);
	V_CalcBlend ();

	r_cache_thrash = false;

	c_brush_polys = 0;
	c_alias_polys = 0;

}


void MYgluPerspective( GLdouble fovy, GLdouble aspect,
		     GLdouble zNear, GLdouble zFar )
{
   GLdouble xmin, xmax, ymin, ymax;

   ymax = zNear * tan( fovy * M_PI / 360.0 );
   ymin = -ymax;

   xmin = ymin * aspect;
   xmax = ymax * aspect;

   glFrustum( xmin, xmax, ymin, ymax, zNear, zFar );

   glGetFloatv(GL_PROJECTION_MATRIX, g_projectionMatrix); //rww
}


//rww - based on gil gribb's paper on frustum plane extraction
void GL_BuildFrustumPlanes(void)
{
	float frsMat[16];

	MatrixMultiply4x4(g_projectionMatrix, r_world_matrix, frsMat);

	g_curFrustum[0][0] = frsMat[3]+frsMat[0];
	g_curFrustum[0][1] = frsMat[4+3]+frsMat[4+0];
	g_curFrustum[0][2] = frsMat[8+3]+frsMat[8+0];
	g_curFrustum[0][3] = frsMat[12+3]+frsMat[12+0];
	PlaneNormalize(g_curFrustum[0]);

	g_curFrustum[1][0] = frsMat[3]-frsMat[0];
	g_curFrustum[1][1] = frsMat[4+3]-frsMat[4+0];
	g_curFrustum[1][2] = frsMat[8+3]-frsMat[8+0];
	g_curFrustum[1][3] = frsMat[12+3]-frsMat[12+0];
	PlaneNormalize(g_curFrustum[1]);

	g_curFrustum[2][0] = frsMat[3]+frsMat[1];
	g_curFrustum[2][1] = frsMat[4+3]+frsMat[4+1];
	g_curFrustum[2][2] = frsMat[8+3]+frsMat[8+1];
	g_curFrustum[2][3] = frsMat[12+3]+frsMat[12+1];
	PlaneNormalize(g_curFrustum[2]);

	g_curFrustum[3][0] = frsMat[3]-frsMat[1];
	g_curFrustum[3][1] = frsMat[4+3]-frsMat[4+1];
	g_curFrustum[3][2] = frsMat[8+3]-frsMat[8+1];
	g_curFrustum[3][3] = frsMat[12+3]-frsMat[12+1];
	PlaneNormalize(g_curFrustum[3]);

	g_curFrustum[4][0] = frsMat[3]+frsMat[2];
	g_curFrustum[4][1] = frsMat[4+3]+frsMat[4+2];
	g_curFrustum[4][2] = frsMat[8+3]+frsMat[8+2];
	g_curFrustum[4][3] = frsMat[12+3]+frsMat[12+2];
	PlaneNormalize(g_curFrustum[4]);

	g_curFrustum[5][0] = frsMat[3]-frsMat[2];
	g_curFrustum[5][1] = frsMat[4+3]-frsMat[4+2];
	g_curFrustum[5][2] = frsMat[8+3]-frsMat[8+2];
	g_curFrustum[5][3] = frsMat[12+3]-frsMat[12+2];
	PlaneNormalize(g_curFrustum[5]);
}

/*
=============
R_SetupGL
=============
*/
void R_SetupGL (void)
{
	float	screenaspect;
	extern	int glwidth, glheight;
	int		x, x2, y2, y, w, h;

	if ((!gl_normalMaps.value || !g_canUseTexFP) && gl_flashblend.value != 1.0f)
	{ //rww - since dlights are placed all over now, don't want big ugly slow lightmap ones
		Cvar_SetValue("gl_flashblend", 1.0f); //use silly shell for dlights
	}
	else if ((gl_normalMaps.value && g_canUseTexFP) && gl_flashblend.value != 0.0f)
	{
		Cvar_SetValue("gl_flashblend", 0.0f); //real dlights
	}

	//
	// set up viewpoint
	//
//	glMatrixMode(GL_PROJECTION);
//    glLoadIdentity ();
	x = r_refdef.vrect.x * glwidth/vid.width;
	x2 = (r_refdef.vrect.x + r_refdef.vrect.width) * glwidth/vid.width;
	y = (vid.height-r_refdef.vrect.y) * glheight/vid.height;
	y2 = (vid.height - (r_refdef.vrect.y + r_refdef.vrect.height)) * glheight/vid.height;

	// fudge around because of frac screen scale
	if (x > 0)
		x--;
	if (x2 < glwidth)
		x2++;
	if (y2 < 0)
		y2--;
	if (y < glheight)
		y++;

	w = x2 - x;
	h = y - y2;

	if (envmap)
	{
		x = y2 = 0;
		w = h = 256;
	}

	glViewport (glx + x, gly + y2, w, h);

	if (gl_forceWideAspect.value)
	{ //rww
		screenaspect = 1280.0f/720.0f;
	}
	else
	{
	    screenaspect = (float)r_refdef.vrect.width/r_refdef.vrect.height;
	}
//	yfov = 2*atan((float)r_refdef.vrect.height/r_refdef.vrect.width)*180/M_PI;
//	yfov = (2.0 * tan (scr_fov.value/360*M_PI)) / screenaspect;
//	yfov = 2*atan((float)r_refdef.vrect.height/r_refdef.vrect.width)*(scr_fov.value*2)/M_PI;
//    MYgluPerspective (yfov,  screenaspect,  4,  4096);
	//rww - moved down

	if (mirror)
	{
		if (mirror_plane->normal[2])
			glScalef (1, -1, 1);
		else
			glScalef (-1, 1, 1);
		glCullFace(GL_BACK);
	}
	else
		glCullFace(GL_FRONT);

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

    glRotatef (-90,  1, 0, 0);	    // put Z going up
    glRotatef (90,  0, 0, 1);	    // put Z going up
    glRotatef (-r_refdef.viewangles[2],  1, 0, 0);
    glRotatef (-r_refdef.viewangles[0],  0, 1, 0);
    glRotatef (-r_refdef.viewangles[1],  0, 0, 1);
    glTranslatef (-r_refdef.vieworg[0],  -r_refdef.vieworg[1],  -r_refdef.vieworg[2]);

	glGetFloatv (GL_MODELVIEW_MATRIX, r_world_matrix);

	/*
	if (!g_canUseTexFP || !gl_normalMaps.value)
	{
		g_beginLodCutoff = 32468.0f;
		g_endLodCutoff = 32768.0f;
	}
	else
	*/
	{
		float svFogNear = atof(Info_ValueForKey(cl.serverinfo, "fognear"));
		float svFogFar = atof(Info_ValueForKey(cl.serverinfo, "fogfar"));
		if (!svFogNear && !svFogFar)
		{ //if the server isn't overriding, take from cvars
			svFogNear = gl_fognear.value;
			svFogFar = gl_fogfar.value;
		}

		g_beginLodCutoff = svFogNear;
		g_endLodCutoff = svFogFar;
		/*
		if (g_beginLodCutoff < 0.0f)
		{
			g_beginLodCutoff = 0.0f;
		}
		*/
		//actually, i'll allow this.
		if (g_endLodCutoff < (g_beginLodCutoff+1.0f))
		{
			g_endLodCutoff = (g_beginLodCutoff+1.0f);
		}
	}

	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
	MYgluPerspective (r_refdef.fov_y,  screenaspect,  4,  g_endLodCutoff);

	GL_BuildFrustumPlanes(); //rww

	//rww
    glLoadIdentity ();
    MYgluPerspective (r_refdef.fov_y,  screenaspect,  4,  32768); //actual far clip plane for stencil volumes
	glMatrixMode(GL_MODELVIEW);
	//set up fog parameters
	glHint(GL_FOG_HINT, GL_NICEST);
	glFogf(GL_FOG_MODE, GL_LINEAR);
	glFogf(GL_FOG_START, g_beginLodCutoff);
	glFogf(GL_FOG_END, g_endLodCutoff);
	glFogfv(GL_FOG_COLOR, g_glClearColor);

	//
	// set drawing parms
	//
	if (gl_cull.value)
		glEnable(GL_CULL_FACE);
	else
		glDisable(GL_CULL_FACE);

	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glEnable(GL_DEPTH_TEST);
}

//rww
void R_ClearChains(void)
{
	int lnum, i;
	msurface_t *s;
	texture_t *t;

	for (i=0 ; i<cl.worldmodel->numtextures ; i++)
	{
		t = cl.worldmodel->textures[i];
		if (!t)
			continue;
		s = t->texturechain;
		if (!s)
			continue;

		t->texturechain = NULL;
	}
}

//rww - draw dynamic light passes
extern qboolean GL_DrawDLightSurfs(int chaincount, int lightNum);
extern cvar_t gl_lightCounts;
void R_RenderFPLights(void)
{
	int lnum;
	int lightsDrawn = 0;

	if (!g_canUseTexFP || !gl_normalMaps.value)
	{ //rww - draw dynamic light passes
		return;
	}

	glDepthMask(GL_FALSE); //light pass needs no depth mask
	for (lnum=0 ; lnum<g_numLiveLights ; lnum++)
	{
		dlight_t *dl = g_liveLights[lnum];
		if (!dl->radius)
			continue;

		if (dl->maplight)
		{
			if (!gl_dynamicMapLights.value)
			{ //only render maplights if the option to do so is set
				continue;
			}
			/*
			if (dl->mapEnt && dl->mapEnt->visframe != r_framecount)
			{ //the associated entity wasn't added this frame
				continue;
			}
			*/
		}

		g_chaincount++;
		if (g_chaincount == 0)
		{ //overflow
			g_chaincount = 1;
		}

		if (GL_DrawDLightSurfs(g_chaincount, g_liveLightsIdx[lnum]))
		{
			lightsDrawn++;
		}
	}

	if (gl_lightCounts.value)
	{ //print how many lights made it through this frame
		Con_Printf("%i lights drawn\n", lightsDrawn);
	}

	glDepthMask(GL_TRUE);
}

//rww
static int GL_SurfSpriteSort(const void *arg1, const void *arg2)
{
	surfSpriteDraw_t *a1 = (surfSpriteDraw_t *)arg1;
	surfSpriteDraw_t *a2 = (surfSpriteDraw_t *)arg2;
	if (a1->viewLen < a2->viewLen)
	{
		return 1;
	}
	else if (a2->viewLen < a1->viewLen)
	{
		return -1;
	}

	return 0;
}

//#define _SOFTWARE_BATCHED_SURFSPRITES

#ifndef _SOFTWARE_BATCHED_SURFSPRITES
#define bglColor4f		glColor4f
#define bglTexCoord2f	glTexCoord2f
#define bglVertex3f		glVertex3f
#define bglBegin		glBegin
#define bglEnd			glEnd
#else
#define bglBegin(a)
#define bglEnd()

#define MAX_SURF_SPR_VERTS 8192
static float surfSprClr[4*MAX_SURF_SPR_VERTS];
static float surfSprTC[2*MAX_SURF_SPR_VERTS];
static float surfSprVerts[3*MAX_SURF_SPR_VERTS];
int surfCurVert = 0;
static int surfQuads[4*(MAX_SURF_SPR_VERTS/4)];
int surfCurQuad = 0;

void RenderSpriteBatch(void)
{
	if (surfCurQuad <= 0)
	{
		surfCurVert = 0;
		surfCurQuad = 0;
		return;
	}

	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FLOAT, 3 * sizeof(GL_FLOAT), surfSprVerts);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer(2, GL_FLOAT, 2 * sizeof(GL_FLOAT), surfSprTC);
	glEnableClientState(GL_COLOR_ARRAY);
	glColorPointer(4, GL_FLOAT, 4 * sizeof(GL_FLOAT), surfSprClr);

	glDrawElements(GL_QUADS, surfCurQuad*4, GL_UNSIGNED_INT, surfQuads);

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);

}

static __inline void bglColor4f(float a, float b, float c, float d)
{
	surfSprClr[(surfCurVert*4)+0] = a;
	surfSprClr[(surfCurVert*4)+1] = b;
	surfSprClr[(surfCurVert*4)+2] = c;
	surfSprClr[(surfCurVert*4)+3] = d;
}
static __inline void bglTexCoord2f(float a, float b)
{
	surfSprTC[(surfCurVert*2)+0] = a;
	surfSprTC[(surfCurVert*2)+1] = b;
}
static __inline void bglVertex3f(float a, float b, float c)
{
	if (surfCurVert >= MAX_SURF_SPR_VERTS)
	{
		return;
	}

	surfSprVerts[(surfCurVert*3)+0] = a;
	surfSprVerts[(surfCurVert*3)+1] = b;
	surfSprVerts[(surfCurVert*3)+2] = c;
	surfCurVert++;
	if ((surfCurVert%4) == 0)
	{
		surfQuads[(surfCurQuad*4)+0] = surfCurQuad-4;
		surfQuads[(surfCurQuad*4)+1] = surfCurQuad-3;
		surfQuads[(surfCurQuad*4)+2] = surfCurQuad-2;
		surfQuads[(surfCurQuad*4)+3] = surfCurQuad-1;
	}
}
#endif

//rww
static __inline void GL_SurfSpriteColor(float *clr, float alpha, float f)
{
	if (g_depthPass)
	{
		return;
	}
	bglColor4f(clr[0]*f, clr[1]*f, clr[2]*f, alpha);
}

//rww
static __inline qboolean GL_SurfForSpriteLitByLight(msurface_t *surf, int lnum)
{
	if ( !(surf->dlightbits & (1<<(lnum%32)) ) )
		return false;

	if ( !(surf->dlightbitsHigh & (1<<(lnum>>5)) ) )
		return false;

	return true;
}

//rww
static __inline qboolean R_SpriteCullBox (vec3_t mins, vec3_t maxs)
{
	int		i;
	float	side;

	for (i=0 ; i<4 ; i++)
		if (BoxOnPlaneSide (mins, maxs, &frustum[i]) == 2)
			return true;

	return false;
}

//rww
void GL_RenderSurfSprite(msurface_t *surf, surfSprite_t *spr, float *mat, float alpha,
						 qboolean doShadow, float viewLen, float frontRange)
{
	float w = spr->baseW;
	float h = spr->baseH;
	float framefactor = 0.9f;//(float)host_frametime * 10.0f;
	float sprMin[3], sprMax[3];
	float clr[3];
	float windIdle[2];
	float vertClr[4] = {0.9f, 0.8f, 0.85f, 0.95f};
	float xNegColor, xPosColor, yNegColor, yPosColor;

	if (!g_depthPass)
	{ //calculate the wind offset based on velocity and sway
		int i;

		windIdle[0] = sinf(cl.time)*0.5f;
		windIdle[1] = cosf(cl.time)*0.5f;

		spr->windVelocity[0] += windIdle[0]*framefactor;
		spr->windVelocity[1] += windIdle[1]*framefactor;

		spr->windEffect[0] += spr->windVelocity[0]*(0.25f*framefactor);
		spr->windEffect[1] += spr->windVelocity[1]*(0.25f*framefactor);
		spr->windVelocity[0] *= (0.9f);
		spr->windVelocity[1] *= (0.9f);

		spr->windVelocity[0] -= spr->windEffect[0]*(0.2f*framefactor);
		spr->windVelocity[1] -= spr->windEffect[1]*(0.2f*framefactor);

		//impose some sane limits on wind effect
		for (i = 0; i < 2; i++)
		{
			if (spr->windEffect[i] > 60.0f)
			{
				spr->windEffect[i] = 60.0f;
				spr->windVelocity[i] = 0.0f;
			}
			else if (spr->windEffect[i] < -60.0f)
			{
				spr->windEffect[i] = -60.0f;
				spr->windVelocity[i] = 0.0f;
			}
		}
	}

	//get some color offsets based on the rotation so the angle makes a little color difference
	xNegColor = 0.5f + (-sinf(spr->offRot))*0.5f;
	xPosColor = 0.5f + sinf(spr->offRot)*0.5f;
	yNegColor = 0.5f + (-cosf(spr->offRot))*0.5f;
	yPosColor = 0.5f + cosf(spr->offRot)*0.5f;

	GL_Bind(spr->texture);

	if (!g_depthPass)
	{
		int lnum;
		vec3_t dist;
		float add;
		float transInvMat[16];
		qboolean hasTransInvMat = false;

		clr[0] = spr->baseColor;
		/*
		clr[0] = ((float)R_LightPoint(spr->basePos) / 255.0f) * 0.45f;
		if (!clr[0])
		{ //don't allow pitch-black
			clr[0] = 0.45f;
		}
		*/
		clr[1] = clr[2] = clr[0];

		for (lnum=0 ; lnum<g_numLiveLights ; lnum++)
		{
			if (!gl_flashblend.value && !GL_SurfForSpriteLitByLight(surf, g_liveLightsIdx[lnum]))
			{ //the surface the sprite is on is not lit (can only use this test with no flashblend)
				continue;
			}
			VectorSubtract (spr->basePos,
							g_liveLights[lnum]->origin,
							dist);
			add = Length(dist)/g_liveLights[lnum]->radius;

			if (add < 1.0f) {
				if (add < 0.0f)
				{
					add = 0.0f;
				}
				add = 1.0f-add;
				if (g_liveLights[lnum]->spriteEff)
				{ //hack - the light's spriteEff flag is set, so blow the sprites around
					float lightLocal[3];

					if (!hasTransInvMat)
					{ //get the transform matrix and invert it, only do this once and not per-light
						float transMat[16];

						glPushMatrix();
						glLoadIdentity();
						glTranslatef(spr->basePos[0], spr->basePos[1], spr->basePos[2]);
						glMultMatrixf(mat);
						glRotatef(spr->offRot, 0, 0, 1);
						glGetFloatv(GL_MODELVIEW_MATRIX, transMat);
						glPopMatrix();

						MatrixInverse4x4(transMat, transInvMat);
						hasTransInvMat = true;
					}

					//get the local light pos and apply to the local velocity
					TransformPointByMatrix4x4(transInvMat, g_liveLights[lnum]->origin, lightLocal);

					spr->windVelocity[0] -= lightLocal[0]*add*0.05f;
					spr->windVelocity[1] -= lightLocal[1]*add*0.05f;
				}
				clr[0] += g_liveLights[lnum]->color[0]*add;
				clr[1] += g_liveLights[lnum]->color[1]*add;
				clr[2] += g_liveLights[lnum]->color[2]*add;

				if (clr[0] > 1.0f ||
					clr[1] > 1.0f ||
					clr[2] > 1.0f)
				{
					int x = 1;
					float c = clr[0];
					float mul;
					while (x < 3)
					{
						if (clr[x] > c)
						{
							c = clr[x];
						}
						x++;
					}
					mul = 1.0f/c;
					clr[0] *= mul;
					clr[1] *= mul;
					clr[2] *= mul;
				}
			}
		}
	}

	sprMin[0] = spr->basePos[0]-24.0f;
	sprMin[1] = spr->basePos[1]-24.0f;
	sprMin[2] = spr->basePos[2]-8.0f;
	sprMax[0] = spr->basePos[0]+24.0f;
	sprMax[1] = spr->basePos[1]+24.0f;
	sprMax[2] = spr->basePos[2]+48.0f;

	if (R_SpriteCullBox(sprMin, sprMax))
	{ //culled, don't draw any of it
		return;
	}

	glPushMatrix();

	glTranslatef(spr->basePos[0], spr->basePos[1], spr->basePos[2]);
	glMultMatrixf(mat);
	glRotatef(spr->offRot, 0, 0, 1);

	if (g_depthPass)
	{ //depth pass only draws the first couple faces
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		bglBegin(GL_QUADS);
			//c1===============================================================
			bglTexCoord2f(0.0f, 1.0f); 
			bglVertex3f(-w*0.5f, 0.0f, 0.0f);
			bglTexCoord2f(0.0f, 0.0f);
			bglVertex3f(-w*0.5f+spr->windEffect[0], 0.0f+spr->windEffect[1], h);
			bglTexCoord2f(1.0f, 0.0f);
			bglVertex3f(w*0.5f+spr->windEffect[0], 0.0f+spr->windEffect[1], h);
			bglTexCoord2f(1.0f, 1.0f);
			bglVertex3f(w*0.5f, 0.0f, 0.0f);
			//c2===============================================================
			bglTexCoord2f(0.0f, 1.0f); 
			bglVertex3f(0.0f, -w*0.5f, 0.0f);
			bglTexCoord2f(0.0f, 0.0f);
			bglVertex3f(0.0f+spr->windEffect[0], -w*0.5f+spr->windEffect[1], h);
			bglTexCoord2f(1.0f, 0.0f);
			bglVertex3f(0.0f+spr->windEffect[0], w*0.5f+spr->windEffect[1], h);
			bglTexCoord2f(1.0f, 1.0f);
			bglVertex3f(0.0f, w*0.5f, 0.0f);
		glEnd();
		glPopMatrix();
		return;
	}

	bglBegin(GL_QUADS);
		//c1===============================================================
		GL_SurfSpriteColor(clr, alpha, vertClr[0]*xNegColor);
		bglTexCoord2f(0.0f, 1.0f); 
		bglVertex3f(-w*0.5f, 0.0f, 0.0f);
		GL_SurfSpriteColor(clr, alpha, vertClr[1]*xNegColor);
		bglTexCoord2f(0.0f, 0.0f);
		bglVertex3f(-w*0.5f+spr->windEffect[0], 0.0f+spr->windEffect[1], h);
		GL_SurfSpriteColor(clr, alpha, vertClr[2]*xPosColor);
		bglTexCoord2f(1.0f, 0.0f);
		bglVertex3f(w*0.5f+spr->windEffect[0], 0.0f+spr->windEffect[1], h);
		GL_SurfSpriteColor(clr, alpha, vertClr[3]*xPosColor);
		bglTexCoord2f(1.0f, 1.0f);
		bglVertex3f(w*0.5f, 0.0f, 0.0f);
		//c2===============================================================
		GL_SurfSpriteColor(clr, alpha, vertClr[0]*yNegColor);
		bglTexCoord2f(0.0f, 1.0f); 
		bglVertex3f(0.0f, -w*0.5f, 0.0f);
		GL_SurfSpriteColor(clr, alpha, vertClr[1]*yNegColor);
		bglTexCoord2f(0.0f, 0.0f);
		bglVertex3f(0.0f+spr->windEffect[0], -w*0.5f+spr->windEffect[1], h);
		GL_SurfSpriteColor(clr, alpha, vertClr[2]*yPosColor);
		bglTexCoord2f(1.0f, 0.0f);
		bglVertex3f(0.0f+spr->windEffect[0], w*0.5f+spr->windEffect[1], h);
		GL_SurfSpriteColor(clr, alpha, vertClr[3]*yPosColor);
		bglTexCoord2f(1.0f, 1.0f);
		bglVertex3f(0.0f, w*0.5f, 0.0f);
		if (viewLen < frontRange*0.75f)
		{
			//l1===============================================================
			GL_SurfSpriteColor(clr, alpha, vertClr[0]*yNegColor);
			bglTexCoord2f(0.0f, 1.0f); 
			bglVertex3f(0.0f, -w*0.5f, 0.0f);
			GL_SurfSpriteColor(clr, alpha, vertClr[1]*yNegColor);
			bglTexCoord2f(0.0f, 0.0f);
			bglVertex3f((h*0.5f)+spr->windEffect[0], -w*0.5f+spr->windEffect[1], h*0.5f);
			GL_SurfSpriteColor(clr, alpha, vertClr[2]*yPosColor);
			bglTexCoord2f(1.0f, 0.0f);
			bglVertex3f((h*0.5f)+spr->windEffect[0], w*0.5f+spr->windEffect[1], h*0.5f);
			GL_SurfSpriteColor(clr, alpha, vertClr[3]*yPosColor);
			bglTexCoord2f(1.0f, 1.0f);
			bglVertex3f(0.0f, w*0.5f, 0.0f);
			//l2===============================================================
			GL_SurfSpriteColor(clr, alpha, vertClr[0]*xNegColor);
			bglTexCoord2f(0.0f, 1.0f); 
			bglVertex3f(-w*0.5f, 0.0f, 0.0f);
			GL_SurfSpriteColor(clr, alpha, vertClr[1]*xNegColor);
			bglTexCoord2f(0.0f, 0.0f);
			bglVertex3f(-w*0.5f+spr->windEffect[0], (h*0.5f)+spr->windEffect[1], h*0.5f);
			GL_SurfSpriteColor(clr, alpha, vertClr[2]*xPosColor);
			bglTexCoord2f(1.0f, 0.0f);
			bglVertex3f(w*0.5f+spr->windEffect[0], (h*0.5f)+spr->windEffect[1], h*0.5f);
			GL_SurfSpriteColor(clr, alpha, vertClr[3]*xPosColor);
			bglTexCoord2f(1.0f, 1.0f);
			bglVertex3f(w*0.5f, 0.0f, 0.0f);
			//l3===============================================================
			GL_SurfSpriteColor(clr, alpha, vertClr[0]*yNegColor);
			bglTexCoord2f(0.0f, 1.0f); 
			bglVertex3f(0.0f, -w*0.5f, 0.0f);
			GL_SurfSpriteColor(clr, alpha, vertClr[1]*yNegColor);
			bglTexCoord2f(0.0f, 0.0f);
			bglVertex3f((-h*0.5f)+spr->windEffect[0], -w*0.5f+spr->windEffect[1], h*0.5f);
			GL_SurfSpriteColor(clr, alpha, vertClr[2]*yPosColor);
			bglTexCoord2f(1.0f, 0.0f);
			bglVertex3f((-h*0.5f)+spr->windEffect[0], w*0.5f+spr->windEffect[1], h*0.5f);
			GL_SurfSpriteColor(clr, alpha, vertClr[3]*yPosColor);
			bglTexCoord2f(1.0f, 1.0f);
			bglVertex3f(0.0f, w*0.5f, 0.0f);
			//l4===============================================================
			GL_SurfSpriteColor(clr, alpha, vertClr[0]*xNegColor);
			bglTexCoord2f(0.0f, 1.0f); 
			bglVertex3f(-w*0.5f, 0.0f, 0.0f);
			GL_SurfSpriteColor(clr, alpha, vertClr[1]*xNegColor);
			bglTexCoord2f(0.0f, 0.0f);
			bglVertex3f(-w*0.5f+spr->windEffect[0], (-h*0.5f)+spr->windEffect[1], h*0.5f);
			GL_SurfSpriteColor(clr, alpha, vertClr[2]*xPosColor);
			bglTexCoord2f(1.0f, 0.0f);
			bglVertex3f(w*0.5f+spr->windEffect[0], (-h*0.5f)+spr->windEffect[1], h*0.5f);
			GL_SurfSpriteColor(clr, alpha, vertClr[3]*xPosColor);
			bglTexCoord2f(1.0f, 1.0f);
			bglVertex3f(w*0.5f, 0.0f, 0.0f);
		}
	bglEnd();

	if (doShadow)
	{ //draw shadow
		GL_Bind(spr->shadowTexture);
		glDepthMask(GL_FALSE);
		glDisable(GL_ALPHA_TEST);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		if (viewLen > (g_beginLodCutoff*0.1f))
		{
			glColor4f(0.0f, 0.0f, 0.0f, (1.0f-((viewLen-(g_beginLodCutoff*0.1f))/((g_beginLodCutoff*0.2f)-(g_beginLodCutoff*0.1f))))*0.5f);
		}
		else
		{
			glColor4f(0.0f, 0.0f, 0.0f, 0.5f);
		}

		w *= 1.5f;
		glBegin(GL_QUADS);
			glTexCoord2f(0.0f, 1.0f); 
			glVertex3f(-w*0.5f+spr->windEffect[0], -w*0.5f+spr->windEffect[1], 0.1f);

			glTexCoord2f(0.0f, 0.0f);
			glVertex3f(-w*0.5f+spr->windEffect[0], w*0.5f+spr->windEffect[1], 0.1f);

			glTexCoord2f(1.0f, 0.0f);
			glVertex3f(w*0.5f+spr->windEffect[0], w*0.5f+spr->windEffect[1], 0.1f);

			glTexCoord2f(1.0f, 1.0f);
			glVertex3f(w*0.5f+spr->windEffect[0], -w*0.5f+spr->windEffect[1], 0.1f);
		glEnd();

		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDisable(GL_BLEND);
		glEnable(GL_ALPHA_TEST);
		glDepthMask(GL_TRUE);
	}

	glPopMatrix();
}

//rww
void GL_RenderSurfSprites(void)
{
	int i;
	//float backRange = g_endLodCutoff*0.25f;
	//float frontRange = g_beginLodCutoff*0.25f;
	float backRange = g_beginLodCutoff*0.25f;
	float frontRange = backRange*0.9f;
	qboolean blending = true;
	static int sortFrame = -1;

	if (g_numSurfSpriteDraws <= 0)
	{ //nothing to draw this frame
		return;
	}

	//sort the surface sprite draw objects
	if (sortFrame != r_framecount)
	{
		qsort(g_surfSpriteDraws, g_numSurfSpriteDraws, sizeof(surfSpriteDraw_t), GL_SurfSpriteSort);
		sortFrame = r_framecount;
	}

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glShadeModel(GL_SMOOTH);

	glDisable(GL_CULL_FACE);
	if (!g_depthPass)
	{
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDepthMask(GL_FALSE);
	}
	else
	{ //depth pass, don't blend
		blending = false;
		glEnable(GL_ALPHA_TEST);
	}

	//go through and draw them all, the ones in the fade range will be blended,
	//while close-up ones will be strictly alpha tested
	for (i = 0; i < g_numSurfSpriteDraws; i++)
	{
		float alpha = 1.0f;
		surfSpriteDraw_t *d = &g_surfSpriteDraws[i];
		if (d->viewLen >= backRange)
		{ //too far away
			continue;
		}

		if (d->viewLen < frontRange && blending)
		{
#ifdef _SOFTWARE_BATCHED_SURFSPRITES
			RenderSpriteBatch();
#endif

			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glDisable(GL_BLEND);
			glEnable(GL_ALPHA_TEST);
			glDepthMask(GL_TRUE);
			blending = false;
		}
		if (blending)
		{ //set the fade based on the distance from the camera
			alpha = 1.0f-((d->viewLen-frontRange)/(backRange-frontRange));
		}
		GL_RenderSurfSprite(d->surf, d->spr, g_surfSpriteDraws[i].mat, alpha, !blending && d->viewLen < (g_beginLodCutoff*0.2f), d->viewLen, frontRange);
	}

#ifdef _SOFTWARE_BATCHED_SURFSPRITES
	RenderSpriteBatch();
#endif

	glEnable(GL_CULL_FACE);
	glDisable(GL_ALPHA_TEST);
	if (blending)
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDisable(GL_BLEND);
		glDepthMask(GL_TRUE);
	}
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	glShadeModel(GL_FLAT);
}

//rww
extern int hr_sun; //rww
extern int hr_sunring; //rww
float Blood_GetMappedValue(float perc, float *values, int numValues); //rww
void R_DrawSunSky(qboolean occlude)
{
	float w = 512.0f;
	float h = 512.0f;
	float d = 8192.0f;
	double corner1 = realtime*2;
	double corner2 = realtime*3;
	double corner3 = realtime*4;
	double corner4 = realtime*5;
	float cornerFactor = 2.0f;
	float n[3];
	int occQuery;
	int occSamples = 1;
	qboolean onePopped = false;
	float occlStep = host_frametime*8.0f;
	static float idealOcclude = 0.0f;
	static float lastOcclude = 0.0f;

	if (occlude)
	{
		g_lightExposureFactor = 0.0f;
	}

	if (!g_hasSun || !hr_sun)
	{
		return;
	}

	VectorSubtract(g_sunPos, r_refdef.vieworg, n);
	VectorNormalize(n);
	n[0] *= d;
	n[1] *= d;
	n[2] *= d;

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glEnable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);

	glDepthMask(GL_FALSE);
	glPushMatrix();

	glTranslatef(r_refdef.vieworg[0], r_refdef.vieworg[1], r_refdef.vieworg[2]);
	glTranslatef(n[0], n[1], n[2]);
	glRotatef (-r_refdef.viewangles[1],  0, 0, -1);
	glRotatef (-r_refdef.viewangles[0],  0, -1, 0);
	glRotatef (-r_refdef.viewangles[2],  -1, 0, 0);
	glPushMatrix();

	glRotatef (-90.0f,  0, 1, 0);
	glRotatef (-90.0f,  0, 0, 1);

	GL_Bind(hr_sun);
	glColor4f(1.0f, 1.0f, 0.7f, 1.0f);

	if (g_canUseOccQuery && occlude)
	{
		int available = 0;
        g_glExt.glGenQueriesARB(1, &occQuery);

        glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
		glDisable(GL_TEXTURE_2D);

        g_glExt.glBeginQueryARB(GL_SAMPLES_PASSED_ARB, occQuery);
		glBegin(GL_QUADS);
			glVertex2f(-(w*0.5f)+(sinf(corner1)*cornerFactor), -(h*0.5f)+(cosf(corner1)*cornerFactor));
			glVertex2f(-(w*0.5f)+(sinf(corner2)*cornerFactor), (h*0.5f)+(cosf(corner2)*cornerFactor));
			glVertex2f((w*0.5f)+(sinf(corner3)*cornerFactor), (h*0.5f)+(cosf(corner3)*cornerFactor));
			glVertex2f((w*0.5f)+(sinf(corner4)*cornerFactor), -(h*0.5f)+(cosf(corner4)*cornerFactor));
		glEnd();
        g_glExt.glEndQueryARB(GL_SAMPLES_PASSED_ARB);

        glFlush();

		while (!available)
		{
			g_glExt.glGetQueryObjectivARB(occQuery, GL_QUERY_RESULT_AVAILABLE_ARB, &available);
		}

		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
		glEnable(GL_TEXTURE_2D);

		occSamples = 0;
		g_glExt.glGetQueryObjectivARB(occQuery, GL_QUERY_RESULT_ARB, &occSamples);
		if (occSamples > 0)
		{ //add lighting exposure from the sun
			float dv[3];
			float lightFactor;
			float fwd[3];
			float frustRad;
			float w = 640.0f / (float)window_width;
			float h = 480.0f / (float)window_height;
			w *= h;

			VectorSubtract(r_refdef.vieworg, g_sunPos, dv);

			//basis for light bloom
			lightFactor = ((float)occSamples * w) / 400.0f;
			AngleVectors(r_refdef.viewangles, fwd, 0, 0);
			VectorNormalize(dv);
			frustRad = DotProduct(fwd, dv);

			if (frustRad <= 0.0f)
			{
				float f = -(frustRad + 0.6f);
				idealOcclude = 0.6f + ((lightFactor*0.3f)*(f*7.0f))*0.4f;
			}
			else
			{
				idealOcclude = 0.6f + ((lightFactor*0.3f)*0.5f)*0.4f;
			}
			if (idealOcclude > 1.0f)
			{
				idealOcclude = 1.0f;
			}

			if (frustRad < -0.6f)
			{
				frustRad += 0.6f;
				frustRad *= 2.5f;
				frustRad = -frustRad;
				/*
				if (frustRad > 0.9f)
				{
					g_lightExposureFactor += ((frustRad*(1.0f-(1.0f/(1024.0f*1.5f))))*lightFactor)*0.75f;
				}
				else
				{
					g_lightExposureFactor += ((frustRad*(1.0f-(1.0f/(1024.0f*1.5f))))*lightFactor)*0.5f;
				}
				*/
				g_lightExposureFactor += ((frustRad*(1.0f-(1.0f/(1024.0f*1.5f))))*lightFactor)*0.6f;
			}
		}
		else
		{
			idealOcclude = 0.0f;
		}
	}

	if (lastOcclude < idealOcclude)
	{
		lastOcclude += occlStep;
		if (lastOcclude > idealOcclude)
		{
			lastOcclude = idealOcclude;
		}
	}
	else if (lastOcclude > idealOcclude)
	{
		lastOcclude -= occlStep;
		if (lastOcclude < idealOcclude)
		{
			lastOcclude = idealOcclude;
		}
	}

	if (occSamples > 0)
	{
		glBegin(GL_QUADS);
			glTexCoord2f(0.0f, 1.0f);
			glVertex2f(-(w*0.5f)+(sinf(corner1)*cornerFactor), -(h*0.5f)+(cosf(corner1)*cornerFactor));
			glTexCoord2f(0.0f, 0.0f);
			glVertex2f(-(w*0.5f)+(sinf(corner2)*cornerFactor), (h*0.5f)+(cosf(corner2)*cornerFactor));
			glTexCoord2f(1.0f, 0.0f);
			glVertex2f((w*0.5f)+(sinf(corner3)*cornerFactor), (h*0.5f)+(cosf(corner3)*cornerFactor));
			glTexCoord2f(1.0f, 1.0f);
			glVertex2f((w*0.5f)+(sinf(corner4)*cornerFactor), -(h*0.5f)+(cosf(corner4)*cornerFactor));
		glEnd();
		if (g_bloomin && lastOcclude > 0.0f)
		{
			float interp[8] = {0.0f, 0.05f, 0.1f, 0.15f, 0.2f, 0.4f, 0.7f, 1.0f};
			float clr = 0.05f;
			float f = Blood_GetMappedValue(lastOcclude, interp, 8);
			w *= (f*24.0f);
			h *= (f*24.0f);

			glPopMatrix();
			onePopped = true;
			glRotatef (f*90.0f,  1, 0, 0);
			glRotatef (-90.0f,  0, 1, 0);
			glRotatef (-90.0f,  0, 0, 1);

			GL_Bind(hr_sunring);
			glColor4f(clr, clr, clr, 1.0f);
			glBlendFunc(GL_ONE, GL_ONE);
			glBegin(GL_QUADS);
				glTexCoord2f(0.0f, 1.0f);
				glVertex2f(-(w*0.5f), -(h*0.5f));
				glTexCoord2f(0.0f, 0.0f);
				glVertex2f(-(w*0.5f), (h*0.5f));
				glTexCoord2f(1.0f, 0.0f);
				glVertex2f((w*0.5f), (h*0.5f));
				glTexCoord2f(1.0f, 1.0f);
				glVertex2f((w*0.5f), -(h*0.5f));
			glEnd();
		}
	}

	if (!onePopped)
	{
		glPopMatrix();
	}
	glPopMatrix();
	glDepthMask(GL_TRUE);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_BLEND);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}

//rww
dlight_t *g_liveLights[MAX_DLIGHTS];
int g_liveLightsIdx[MAX_DLIGHTS];
int g_numLiveLights = 0;

//rww
void R_BuildLiveLightList(void)
{
	int lnum;
	g_numLiveLights = 0;
	for (lnum = 0; lnum < MAX_DLIGHTS; lnum++)
	{
		if (cl_dlights[lnum].die >= cl.time)
		{
			g_liveLights[g_numLiveLights] = &cl_dlights[lnum];
			g_liveLightsIdx[g_numLiveLights] = lnum;
			g_numLiveLights++;
		}
	}
}

//rww
extern float g_breathAmt;
extern int screen_accum_textures[NUM_ACCUM_TEXTURES];
void GL_BreathPass(void)
{
	static int builtFrames = 0;
	static int lastAccum = 0;
	static double capTime = 0.0;
	float a;
	int i;
	int tex;
	if (!g_canUseTexRect)
	{
		return;
	}

	if (g_breathAmt == 0.0f)
	{
		builtFrames = 0;
		lastAccum = 0;
		capTime = 0.0;
		return;
	}

	if (builtFrames < NUM_ACCUM_TEXTURES)
	{
		builtFrames++;
	}

	if (capTime < realtime)
	{
		float d;
		lastAccum = (lastAccum+1)%NUM_ACCUM_TEXTURES;
		glDisable(GL_TEXTURE_2D);
		glEnable(GL_TEXTURE_RECTANGLE_NV); 
		glBindTexture(GL_TEXTURE_RECTANGLE_NV, screen_accum_textures[lastAccum]);
		glCopyTexSubImage2D(GL_TEXTURE_RECTANGLE_NV, 0, 0, 0, 0, 0, window_width, window_height); 
		glDisable(GL_TEXTURE_RECTANGLE_NV);
		glEnable(GL_TEXTURE_2D);
		//capTime = realtime + 0.05;
	}

	tex = lastAccum-builtFrames;
	if (tex < 0)
	{
		tex += NUM_ACCUM_TEXTURES;
	}
	a = 1.0f-(g_breathAmt*0.12f);
	if (a < 0.25f)
	{
		a = 0.25f;
	}
	glEnable(GL_BLEND);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glColor4f(1.0f, 1.0f, 1.0f, a);
	for (i = 0; i < builtFrames-1; i++)
	{
		tex++;
		if (tex >= NUM_ACCUM_TEXTURES)
		{
			tex = 0;
		}
		GL_FullscreenQuad(screen_accum_textures[tex], window_width, window_height, true);
	}
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDisable(GL_BLEND);
}

void R_DrawBotWP(void);

/*
================
R_RenderScene

r_refdef must be set before the first call
================
*/
void Decal_CheckDeferred(void); //rww
void Decal_RenderClientDecals(void); //rww
void Blood_RenderBlobs(void); //rww
extern cvar_t gl_showTris; //rww
void R_RenderScene (void)
{
	qboolean didBloom = false;

	R_BuildLiveLightList(); //rww

	R_SetupFrame ();

	R_SetFrustum ();

	R_SetupGL ();

	R_MarkLeaves ();	// done here so we know if we're in water

	//rww - enable fog for regular passes
	glEnable(GL_FOG);

	R_DrawWorld ();		// adds static entities to the list

	S_ExtraUpdate ();	// don't let sound get messed up if going slow

	Decal_CheckDeferred(); //rww

	R_DrawEntitiesOnList ();

	R_DrawViewModel (-1); //rww

	R_DrawWaterSurfaces(); //rww - moved here

	GL_RenderSurfSprites(); //rww

	R_DrawEntitySunVolumes(); //rww

	glDisable(GL_FOG); //rww

	R_RenderFPLights(); //rww

	Decal_RenderClientDecals(); //rww

	GL_DisableMultitexture();

	R_RenderDlights ();

	glEnable(GL_FOG); //rww
	R_DrawTreeLeafs(); //rww
	R_DrawParticles ();
	Blood_RenderBlobs(); //rww

	glDisable(GL_FOG); //rww

	if (gl_showTris.value)
	{ //rww
		glDisable(GL_TEXTURE_2D);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glDepthMask(GL_FALSE);
		glLineWidth(2.0f);
		g_depthPass = true;
		DrawTextureChains();
		R_DrawEntitiesOnList();
		GL_RenderSurfSprites();
		g_depthPass = false;
		glLineWidth(1.0f);
		glDepthMask(GL_TRUE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glEnable(GL_TEXTURE_2D);
	}

	R_DrawBotWP(); //rww - shouldn't do anything most of the time

	if (gl_bloom.value)
	{ //rww
		didBloom = GL_BloomPass();
	}
	if (!didBloom && g_canUseTexRect)
	{ //rww - hack to keep saturation level the same
		float f = 1.0f;
		GL_CapScreenTex();
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE, GL_ONE);

		glColor4f(f, f, f, 1.0f);
		GL_FullscreenQuad(screen_texture, window_width, window_height, true);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDisable(GL_BLEND);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	}

	GL_BreathPass(); //rww

	R_ClearChains(); //rww

#ifdef GLTEST
	Test_Draw ();
#endif

}


/*
=============
R_Clear
=============
*/
void R_Clear (void)
{
	//rww
	if (g_glClearColor[0] != gl_fog_r.value ||
		g_glClearColor[1] != gl_fog_g.value ||
		g_glClearColor[2] != gl_fog_b.value)
	{
		g_glClearColor[0] = gl_fog_r.value;
		g_glClearColor[1] = gl_fog_g.value;
		g_glClearColor[2] = gl_fog_b.value;
		glClearColor (g_glClearColor[0], g_glClearColor[1],
			g_glClearColor[2], g_glClearColor[3]);
	}

	if (r_mirroralpha.value != 1.0)
	{
		if (gl_clear.value)
			glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		else
			glClear (GL_DEPTH_BUFFER_BIT);
		gldepthmin = 0;
		gldepthmax = 0.5;
		glDepthFunc (GL_LEQUAL);
	}
	else if (gl_ztrick.value)
	{
		static int trickframe;

		if (gl_clear.value)
			glClear (GL_COLOR_BUFFER_BIT);

		trickframe++;
		if (trickframe & 1)
		{
			gldepthmin = 0;
			gldepthmax = 0.49999;
			glDepthFunc (GL_LEQUAL);
		}
		else
		{
			gldepthmin = 1;
			gldepthmax = 0.5;
			glDepthFunc (GL_GEQUAL);
		}
	}
	else
	{
		if (gl_clear.value)
			glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		else
			glClear (GL_DEPTH_BUFFER_BIT);
		gldepthmin = 0;
		gldepthmax = 1;
		glDepthFunc (GL_LEQUAL);
	}

	glDepthRange (gldepthmin, gldepthmax);
}

#if 0 //!!! FIXME, Zoid, mirror is disabled for now
/*
=============
R_Mirror
=============
*/
void R_Mirror (void)
{
	float		d;
	msurface_t	*s;
	entity_t	*ent;

	if (!mirror)
		return;

	memcpy (r_base_world_matrix, r_world_matrix, sizeof(r_base_world_matrix));

	d = DotProduct (r_refdef.vieworg, mirror_plane->normal) - mirror_plane->dist;
	VectorMA (r_refdef.vieworg, -2*d, mirror_plane->normal, r_refdef.vieworg);

	d = DotProduct (vpn, mirror_plane->normal);
	VectorMA (vpn, -2*d, mirror_plane->normal, vpn);

	r_refdef.viewangles[0] = -asin (vpn[2])/M_PI*180;
	r_refdef.viewangles[1] = atan2 (vpn[1], vpn[0])/M_PI*180;
	r_refdef.viewangles[2] = -r_refdef.viewangles[2];

	ent = &cl_entities[cl.viewentity];
	if (cl_numvisedicts < MAX_VISEDICTS)
	{
		cl_visedicts[cl_numvisedicts] = ent;
		cl_numvisedicts++;
	}

	gldepthmin = 0.5;
	gldepthmax = 1;
	glDepthRange (gldepthmin, gldepthmax);
	glDepthFunc (GL_LEQUAL);

	R_RenderScene ();
	R_DrawWaterSurfaces ();


	gldepthmin = 0;
	gldepthmax = 0.5;
	glDepthRange (gldepthmin, gldepthmax);
	glDepthFunc (GL_LEQUAL);

	// blend on top
	glEnable (GL_BLEND);
	glMatrixMode(GL_PROJECTION);
	if (mirror_plane->normal[2])
		glScalef (1,-1,1);
	else
		glScalef (-1,1,1);
	glCullFace(GL_FRONT);
	glMatrixMode(GL_MODELVIEW);

	glLoadMatrixf (r_base_world_matrix);

	glColor4f (1,1,1,r_mirroralpha.value);
	s = cl.worldmodel->textures[mirrortexturenum]->texturechain;
	for ( ; s ; s=s->texturechain)
		R_RenderBrushPoly (NULL, s, -1);
	cl.worldmodel->textures[mirrortexturenum]->texturechain = NULL;
	glDisable (GL_BLEND);
	glColor4f (1,1,1,1);
}
#endif

/*
================
R_RenderView

r_refdef must be set before the first call
================
*/
void R_RenderView (void)
{
	double	time1 = 0, time2;

	if (r_norefresh.value)
		return;

	if (!r_worldentity.model || !cl.worldmodel)
		Sys_Error ("R_RenderView: NULL worldmodel");

	if (r_speeds.value)
	{
		glFinish ();
		time1 = Sys_DoubleTime ();
		c_brush_polys = 0;
		c_alias_polys = 0;
	}

	mirror = false;

	if (gl_finish.value)
		glFinish ();

	R_Clear ();

	// render normal view
	R_RenderScene ();
	//R_DrawViewModel (-1); //rww
	//R_DrawWaterSurfaces (); //rww

	// render mirror view
//	R_Mirror ();

	R_PolyBlend ();

	if (r_speeds.value)
	{
//		glFinish ();
		time2 = Sys_DoubleTime ();
		Con_Printf ("%3i ms  %4i wpoly %4i epoly\n", (int)((time2-time1)*1000), c_brush_polys, c_alias_polys); 
	}
}
