/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "quakedef.h"
#include "winquake.h"

extern int hr_plaque; //rww

void (*vid_menudrawfn)(void);
void (*vid_menukeyfn)(int key);

void (*vid_optmenudrawfn)(void); //rww
void (*vid_optmenukeyfn)(int key); //rww

enum {m_none, m_main, m_singleplayer, m_load, m_save, m_multiplayer, m_setup, m_svbrowser, m_infomenu, m_net, m_options, m_video, m_optvideo /*rww*/, m_keys, m_help, m_quit, m_serialconfig, m_modemconfig, m_lanconfig, m_gameoptions, m_search, m_slist} m_state;

void M_DrawHiResTex(int x, int y, int tex, int w, int h);

void M_Menu_Main_f (void);
	void M_Menu_SinglePlayer_f (void);
		void M_Menu_Load_f (void);
		void M_Menu_Save_f (void);
	void M_Menu_MultiPlayer_f (void);
		void M_Menu_Setup_f (void);
		void M_Menu_Net_f (void);
	void M_Menu_Options_f (void);
		void M_Menu_Keys_f (void);
		void M_Menu_Video_f (void);
		void M_Menu_OptVideo_f (void);
	void M_Menu_Help_f (void);
	void M_Menu_Quit_f (void);
void M_Menu_SerialConfig_f (void);
	void M_Menu_ModemConfig_f (void);
void M_Menu_LanConfig_f (void);
void M_Menu_GameOptions_f (void);
void M_Menu_Search_f (void);
void M_Menu_ServerList_f (void);

void M_InfoMenu_Draw(void); //rww

void M_Main_Draw (void);
	void M_SinglePlayer_Draw (void);
		void M_Load_Draw (void);
		void M_Save_Draw (void);
	void M_MultiPlayer_Draw (void);
		void M_Setup_Draw (void);
		void M_SvBrowser_Draw (void); //rww
		void M_Net_Draw (void);
	void M_Options_Draw (void);
		void M_Keys_Draw (void);
		void M_Video_Draw (void);
		void M_OptVideo_Draw (void); //rww
	void M_Help_Draw (void);
	void M_Quit_Draw (void);
void M_SerialConfig_Draw (void);
	void M_ModemConfig_Draw (void);
void M_LanConfig_Draw (void);
void M_GameOptions_Draw (void);
void M_Search_Draw (void);
void M_ServerList_Draw (void);

void M_InfoMenu_Key(int key); //rww

void M_Main_Key (int key);
	void M_SinglePlayer_Key (int key);
		void M_Load_Key (int key);
		void M_Save_Key (int key);
	void M_MultiPlayer_Key (int key);
		void M_Setup_Key (int key);
		void M_SvBrowser_Key (int key); //rww
		void M_Net_Key (int key);
	void M_Options_Key (int key);
		void M_Keys_Key (int key);
		void M_Video_Key (int key);
		void M_OptVideo_Key (int key); //rww
	void M_Help_Key (int key);
	void M_Quit_Key (int key);
void M_SerialConfig_Key (int key);
	void M_ModemConfig_Key (int key);
void M_LanConfig_Key (int key);
void M_GameOptions_Key (int key);
void M_Search_Key (int key);
void M_ServerList_Key (int key);

qboolean	m_entersound;		// play after drawing a frame, so caching
								// won't disrupt the sound
qboolean	m_recursiveDraw;

int			m_return_state;
qboolean	m_return_onerror;
char		m_return_reason [32];

#define StartingGame	(m_multiplayer_cursor == 1)
#define JoiningGame		(m_multiplayer_cursor == 0)
#define SerialConfig	(m_net_cursor == 0)
#define DirectConfig	(m_net_cursor == 1)
#define	IPXConfig		(m_net_cursor == 2)
#define	TCPIPConfig		(m_net_cursor == 3)

void M_ConfigureNetSubsystem(void);

//=============================================================================
/* Support Routines */

/*
================
M_DrawCharacter

Draws one solid graphics character
================
*/
void M_DrawCharacter (int cx, int line, int num)
{
	Draw_Character ( cx + ((vid.width - 320)>>1), line, num);
}

void M_Print (int cx, int cy, char *str)
{
	while (*str)
	{
		M_DrawCharacter (cx, cy, (*str)+128);
		str++;
		cx += 8;
	}
}

void M_PrintWhite (int cx, int cy, char *str)
{
	while (*str)
	{
		M_DrawCharacter (cx, cy, *str);
		str++;
		cx += 8;
	}
}

void M_DrawTransPic (int x, int y, qpic_t *pic)
{
	Draw_TransPic (x + ((vid.width - 320)>>1), y, pic);
}

void M_DrawPic (int x, int y, qpic_t *pic)
{
	Draw_Pic (x + ((vid.width - 320)>>1), y, pic);
}

byte identityTable[256];
byte translationTable[256];

void M_BuildTranslationTable(int top, int bottom)
{
	int		j;
	byte	*dest, *source;

	for (j = 0; j < 256; j++)
		identityTable[j] = j;
	dest = translationTable;
	source = identityTable;
	memcpy (dest, source, 256);

	if (top < 128)	// the artists made some backwards ranges.  sigh.
		memcpy (dest + TOP_RANGE, source + top, 16);
	else
		for (j=0 ; j<16 ; j++)
			dest[TOP_RANGE+j] = source[top+15-j];

	if (bottom < 128)
		memcpy (dest + BOTTOM_RANGE, source + bottom, 16);
	else
		for (j=0 ; j<16 ; j++)
			dest[BOTTOM_RANGE+j] = source[bottom+15-j];		
}


void M_DrawTransPicTranslate (int x, int y, qpic_t *pic)
{
	Draw_TransPicTranslate (x + ((vid.width - 320)>>1), y, pic, translationTable);
}


void M_DrawTextBox (int x, int y, int width, int lines)
{
	qpic_t	*p;
	int		cx, cy;
	int		n;

	// draw left side
	cx = x;
	cy = y;
	p = Draw_CachePic ("gfx/box_tl.lmp");
	M_DrawTransPic (cx, cy, p);
	p = Draw_CachePic ("gfx/box_ml.lmp");
	for (n = 0; n < lines; n++)
	{
		cy += 8;
		M_DrawTransPic (cx, cy, p);
	}
	p = Draw_CachePic ("gfx/box_bl.lmp");
	M_DrawTransPic (cx, cy+8, p);

	// draw middle
	cx += 8;
	while (width > 0)
	{
		cy = y;
		p = Draw_CachePic ("gfx/box_tm.lmp");
		M_DrawTransPic (cx, cy, p);
		p = Draw_CachePic ("gfx/box_mm.lmp");
		for (n = 0; n < lines; n++)
		{
			cy += 8;
			if (n == 1)
				p = Draw_CachePic ("gfx/box_mm2.lmp");
			M_DrawTransPic (cx, cy, p);
		}
		p = Draw_CachePic ("gfx/box_bm.lmp");
		M_DrawTransPic (cx, cy+8, p);
		width -= 2;
		cx += 16;
	}

	// draw right side
	cy = y;
	p = Draw_CachePic ("gfx/box_tr.lmp");
	M_DrawTransPic (cx, cy, p);
	p = Draw_CachePic ("gfx/box_mr.lmp");
	for (n = 0; n < lines; n++)
	{
		cy += 8;
		M_DrawTransPic (cx, cy, p);
	}
	p = Draw_CachePic ("gfx/box_br.lmp");
	M_DrawTransPic (cx, cy+8, p);
}

//=============================================================================

//rww
void M_InfoMenu_f(void)
{
	key_dest = key_menu;
	m_state = m_infomenu;
}

char m_infostr[2048];

//rww
void M_InfoMenu(char *msg)
{
	strcpy(m_infostr, msg);
	M_InfoMenu_f();
}

//rww
void M_InfoMenu_Draw(void)
{
	if (hr_plaque)
	{ //rww
		M_DrawHiResTex(0, 4, hr_plaque, 48, 216);
	}
	else
	{
		M_DrawTransPic (0, 4, Draw_CachePic ("gfx/qplaque.lmp") );
	}


	M_Print(80, 24, "Game ended.");
	M_Print(80, 32, "===========");
	M_Print(80, 48, "Reason:");
	M_Print(80, 56, m_infostr);
	M_Print(88, 72, "ok");
	M_DrawCharacter(72, 72, 12+((int)(realtime*4)&1));
}

//rww
void M_InfoMenu_Key(int key)
{
	if (key == K_ESCAPE || key == K_ENTER)
	{
		S_LocalSound ("misc/menu2.wav");
		M_Menu_Main_f ();
	}
}

int m_save_demonum;
	
/*
================
M_ToggleMenu_f
================
*/
void M_ToggleMenu_f (void)
{
	m_entersound = true;

	if (key_dest == key_menu)
	{
		if (m_state != m_main)
		{
			M_Menu_Main_f ();
			return;
		}
		key_dest = key_game;
		m_state = m_none;
		return;
	}
	if (key_dest == key_console && cls.state != ca_disconnected)
	{ //rww - added ca_disconnected check
		Con_ToggleConsole_f ();
	}
	else
	{
		M_Menu_Main_f ();
	}
}

		
//=============================================================================
/* MAIN MENU */

int	m_main_cursor;
#define	MAIN_ITEMS	3//5 //rww


void M_Menu_Main_f (void)
{
	if (key_dest != key_menu)
	{
		m_save_demonum = cls.demonum;
		cls.demonum = -1;
	}
	key_dest = key_menu;
	m_state = m_main;
	m_entersound = true;
}

//rww
void M_DrawHiResTex(int x, int y, int tex, int w, int h)
{
	float texw = 1.0f;
	float texh = 1.0f;

	x += ((vid.width - 320)>>1);

	glColor4f (1,1,1,1);
	GL_Bind (tex);
	glBegin (GL_QUADS);
		glTexCoord2f(0.0f, 0.0f); 
		glVertex2f((float)x, (float)y);
		glTexCoord2f(0.0f, texh);
		glVertex2f((float)x, (float)y+(float)h);
		glTexCoord2f(texw, texh);
		glVertex2f((float)x+(float)w, (float)y+(float)h);
		glTexCoord2f(texw, 0.0f);
		glVertex2f((float)x+(float)w, (float)y);
	glEnd ();
}

void M_Main_Draw (void)
{
	int		f;
	int		cursoffset;
	qpic_t	*p;

	if (hr_plaque)
	{ //rww
		M_DrawHiResTex(0, 4, hr_plaque, 48, 216);
	}
	else
	{
		M_DrawTransPic (0, 4, Draw_CachePic ("gfx/qplaque.lmp") );
	}
	//rww
	//p = Draw_CachePic ("gfx/ttl_main.lmp");
	//M_DrawPic ( (320-p->width)/2, 4, p);
	M_Print(72, 8, "Main");

	M_DrawTransPic (72, 32, Draw_CachePic ("gfx/mainmenu.lmp") );

	f = (int)(realtime * 10)%6;
	
	cursoffset = m_main_cursor;
	if (cursoffset == 2)
	{ //rww - hax
		cursoffset = 4;
	}
	//M_DrawTransPic (54, 32 + cursoffset * 20,Draw_CachePic( va("gfx/menudot%i.lmp", f+1 ) ) );
	M_DrawCharacter (54, 32 + (cursoffset*20) + 4, 12+((int)(realtime*4)&1));
}


void M_Main_Key (int key)
{
	switch (key)
	{
	case K_ESCAPE:
		key_dest = key_game;
		m_state = m_none;
		cls.demonum = m_save_demonum;
		if (cls.demonum != -1 && !cls.demoplayback && cls.state == ca_disconnected)
			CL_NextDemo ();
		break;
		
	case K_DOWNARROW:
		S_LocalSound ("misc/menu1.wav");
		if (++m_main_cursor >= MAIN_ITEMS)
			m_main_cursor = 0;
		break;

	case K_UPARROW:
		S_LocalSound ("misc/menu1.wav");
		if (--m_main_cursor < 0)
			m_main_cursor = MAIN_ITEMS - 1;
		break;

	case K_ENTER:
		m_entersound = true;

		switch (m_main_cursor)
		{
			/*
		case 0:
			M_Menu_SinglePlayer_f ();
			break;
			*/

		case 0://1:
			M_Menu_MultiPlayer_f ();
			break;

		case 1://2:
			M_Menu_Options_f ();
			break;

			/*
		case 3:
			M_Menu_Help_f ();
			break;
			*/

		case 2://4:
			M_Menu_Quit_f ();
			break;
		}
	}
}

//rww - this is all terrible browser menu stuff that was written in one night while
//terribly deprived of sleep so it's generally sloppy and uncommented. sorry.

#define BROWSER_SORT_PING		0
#define BROWSER_SORT_PLAYERS	1
int browser_sortby = BROWSER_SORT_PING;
int browser_numsv = 0;
int browser_numsorted = 0;
int browser_numqueried = 0;
int browser_numreplied = 0;
int browser_selected = 0;
int browser_svnumbase = 0;
qboolean browser_looking = false;

typedef struct browseSv_s
{
	netadr_t		adr;
	qboolean		valid;
	qboolean		shouldDisplay;
	unsigned long	pingOut;
	int				ping;
	char			hostname[16];
	char			mapname[12];
	int				pl;
	int				maxPl;
} browseSv_t;

#define MAX_BROWSE_SERVERS 1024
browseSv_t g_browseServers[MAX_BROWSE_SERVERS];
browseSv_t *g_sortedServers[MAX_BROWSE_SERVERS];

#define MAX_BROWSER_SV_DISPLAY	15//10
#define NUM_BROWSER_ITEMS		2

char browser_status[512];

//rww
static int M_ServerSort(const void *arg1, const void *arg2)
{
	browseSv_t *a1 = *((browseSv_t **)arg1);
	browseSv_t *a2 = *((browseSv_t **)arg2);
	if (browser_sortby == BROWSER_SORT_PING)
	{
		if (a1->ping > a2->ping)
		{
			return 1;
		}
		else if (a2->ping < a1->ping)
		{
			return -1;
		}
	}
	else
	{
		if (a1->pl < a2->pl)
		{
			return 1;
		}
		else if (a2->pl < a1->pl)
		{
			return -1;
		}
	}

	return -1;//0;
}

//rww
qboolean M_BrowserResponse(void)
{
	int i;
	int c;
	browseSv_t *bsv = NULL;
	qboolean found = false;

	for (i = 0; i < browser_numsv; i++)
	{
		bsv = &g_browseServers[i];
		if (*((int *)bsv->adr.ip) == *((int *)net_from.ip) &&
			bsv->adr.port == net_from.port)
		{
			found = true;
			break;
		}
	}

	if (!found)
	{ //not in the list
		return false;
	}

	MSG_BeginReading();
	c = MSG_ReadByte();
	if (c == A2A_ACK)
	{
		if (!bsv->valid)
		{
			char data[11];

			bsv->ping = (int)(GetTickCount()-bsv->pingOut);
			if (bsv->ping < 0)
			{
				bsv->ping = 0;
			}
			else if (bsv->ping > 9999)
			{
				bsv->ping = 9999;
			}
			bsv->valid = true;
			browser_numreplied++;

			//send off a status request
			bsv->pingOut = GetTickCount();
			data[0] = 0xff;
			data[1] = 0xff;
			data[2] = 0xff;
			data[3] = 0xff;
			strcpy(&data[4], "status");
			data[10] = 0;
			NET_SendPacket (11, data, bsv->adr);
			return true;
		}
	}
	else if (bsv->valid && 
		c == 255 && MSG_ReadByte() == 255 && MSG_ReadByte() == 255 && MSG_ReadByte() == 255)
	{ //connectionless message
		c = MSG_ReadByte();
		if (c == A2C_PRINT)
		{
			char *s = MSG_ReadString();
			char *inf = Info_ValueForKey(s, "hostname");
			if (inf && inf[0])
			{
				for (i = 0; i < 15 && inf[i]; i++)
				{
					bsv->hostname[i] = inf[i];
				}
				bsv->hostname[i] = 0;
			}
			inf = Info_ValueForKey(s, "maxclients");
			if (inf && inf[0])
			{
				bsv->maxPl = atoi(inf);
			}
			//count number of active players based on linebreaks in string (silly)
			bsv->pl = -1;
			for (i = 0; s[i]; i++)
			{
				if (s[i] == '\n')
				{
					bsv->pl++;
				}
			}
			inf = Info_ValueForKey(s, "map");
			if (inf && inf[0])
			{
				for (i = 0; i < 10 && inf[i]; i++)
				{
					bsv->mapname[i] = inf[i];
				}
				bsv->mapname[i] = 0;
			}

			if (!bsv->shouldDisplay)
			{ //add to the sorted list
				g_sortedServers[browser_numsorted++] = bsv;
				bsv->shouldDisplay = true;

				sprintf(browser_status, "Querying servers: %i/%i. Replies: %i.", browser_numqueried, browser_numsv, browser_numsorted);

				//re-sort the list
				qsort(g_sortedServers, browser_numsorted, sizeof(g_sortedServers[0]), M_ServerSort);

				if (browser_numsorted == browser_numsv)
				{ //got everything
					browser_looking = false;
				}
			}
			return true;
		}
	}

	MSG_BeginReading(); //if it wasn't handled, reset the read count
	return false;
}

//rww
void M_BrowserAddServer(byte *ip, int port)
{
	browseSv_t *bsv;

	if (!browser_looking)
	{
		return;
	}

	if (browser_numsv >= MAX_BROWSE_SERVERS)
	{ //i don't anticipate this ever happening
		return;
	}

	bsv = &g_browseServers[browser_numsv];
	memset(bsv, 0, sizeof(browseSv_t));
	browser_numsv++;

	sprintf(browser_status, "Querying servers: %i/%i. Replies: %i.", browser_numqueried, browser_numsv, browser_numsorted);

	bsv->adr.ip[0] = ip[0];	
	bsv->adr.ip[1] = ip[1];
	bsv->adr.ip[2] = ip[2];
	bsv->adr.ip[3] = ip[3];
	bsv->adr.port = port;
}

//rww
void M_QueryServer(browseSv_t *bsv)
{
	char data[6];
	bsv->pingOut = GetTickCount();
	data[0] = 0xff;
	data[1] = 0xff;
	data[2] = 0xff;
	data[3] = 0xff;
	data[4] = A2A_PING;
	data[5] = 0;
	NET_SendPacket (6, data, bsv->adr);
}

//rww
double lastBrowseThink = 0.0;
#define SERVER_QUERIES_PER_TICK		8
void M_BrowserLogic(void)
{
	int i;
	if (lastBrowseThink > realtime)
	{
		return;
	}
	if (!browser_numsv || !browser_looking)
	{
		return;
	}

	lastBrowseThink = realtime + 0.5;
	for (i = 0; i < SERVER_QUERIES_PER_TICK && browser_numqueried < browser_numsv; i++, browser_numqueried++)
	{
		M_QueryServer(&g_browseServers[browser_numqueried]);
	}
	sprintf(browser_status, "Querying servers: %i/%i. Replies: %i.", browser_numqueried, browser_numsv, browser_numsorted);
}

//rww
extern netadr_t cl_master_adr;
void M_Menu_SvBrowser_f(void)
{
	char data[2];

	if (cls.demoplayback)
	{
		Cbuf_AddText("disconnect\n");
	}

	memset(g_browseServers, 0, sizeof(g_browseServers));
	memset(g_sortedServers, 0, sizeof(g_sortedServers));

	key_dest = key_menu;
	m_state = m_svbrowser;

	lastBrowseThink = 0.0;
	browser_numsv = 0;
	browser_numsorted = 0;
	browser_numqueried = 0;
	browser_numreplied = 0;
	browser_selected = 0;
	browser_svnumbase = 0;
	browser_looking = true;

	strcpy(browser_status, "Waiting for master list.");

	//send a request to the qr master
	data[0] = C2M_GETLIST;
	data[1] = 0;
	NET_SendPacket (2, data, cl_master_adr);
}

//rww
void M_SvBrowser_Draw (void)
{
	int i;
	int j;
	int printNum = 0;
	char t[512];
	int cursorDrawOffset = browser_selected;

	M_BrowserLogic();

	if (hr_plaque)
	{
		M_DrawHiResTex(0, 4, hr_plaque, 48, 216);
	}
	else
	{
		M_DrawTransPic (16, 4, Draw_CachePic ("gfx/qplaque.lmp") );
	}

	M_Print(72, 8, "Quake Royale Server Browser");

	M_Print(72, 24, browser_status);

	if (browser_looking)
	{
		M_Print(72, 32, "Stop");
	}
	else
	{
		M_Print(72, 32, "Update");
	}

	if (browser_sortby == BROWSER_SORT_PING)
	{
		M_Print(72, 48, "Sort by: Ping");
	}
	else
	{
		M_Print(72, 48, "Sort by: Players");
	}

	M_PrintWhite(72, 56, "Name");
	M_PrintWhite(200, 56, "Ping");
	M_PrintWhite(240, 56, "Players");
	M_PrintWhite(312, 56, "Map");

	for (i = browser_svnumbase, j = 0; i < browser_numsorted, j < MAX_BROWSER_SV_DISPLAY; i++, j++)
	{
		if (g_sortedServers[i] && g_sortedServers[i]->shouldDisplay)
		{
			browseSv_t *sv = g_sortedServers[i];
			printNum++;

			sprintf(t, "%03i. ", i+1);
			M_Print(16, 56+(8*printNum), t);

			M_Print(72, 56+(8*printNum), sv->hostname);
			sprintf(t, "%i", sv->ping);
			M_Print(200, 56+(8*printNum), t);
			sprintf(t, "%i/%i", sv->pl, sv->maxPl);
			M_Print(240, 56+(8*printNum), t);
			M_Print(312, 56+(8*printNum), sv->mapname);
		}
	}

	M_Print(72, 56+(8*MAX_BROWSER_SV_DISPLAY)+16, "Use left/right arrows to scroll");
	M_Print(72, 56+(8*MAX_BROWSER_SV_DISPLAY)+16+8, "through the server list.");

	if (cursorDrawOffset > 1)
	{
		int servSel = (browser_selected-NUM_BROWSER_ITEMS)+browser_svnumbase;
		if (servSel >= 0 && servSel < browser_numsorted)
		{
			char str[512];
			browseSv_t *sv = g_sortedServers[servSel];
			sprintf(str, "Server: %i.%i.%i.%i:%i", sv->adr.ip[0],
				sv->adr.ip[1], sv->adr.ip[2], sv->adr.ip[3], BigShort(sv->adr.port));

			M_PrintWhite(72, 56+(8*MAX_BROWSER_SV_DISPLAY)+16+8+16, str);
		}

		cursorDrawOffset += 2;
	}
	else if (cursorDrawOffset > 0)
	{
		cursorDrawOffset++;
	}

	M_DrawCharacter (56, 32 + (cursorDrawOffset*8), 12+((int)(realtime*4)&1));
}

int mp_cursor = 0; //rww

//rww
void M_SvBrowser_Key (int key)
{
	int numBrowserItems = NUM_BROWSER_ITEMS;
	if (browser_numsorted < MAX_BROWSER_SV_DISPLAY)
	{
		numBrowserItems += browser_numsorted;
	}
	else
	{
		numBrowserItems += MAX_BROWSER_SV_DISPLAY;
	}
	switch (key)
	{
	case K_ESCAPE:
		S_LocalSound ("misc/menu2.wav");
		lastBrowseThink = 0.0;
		browser_numsv = 0;
		browser_numsorted = 0;
		browser_numqueried = 0;
		browser_numreplied = 0;
		browser_looking = false;
		M_Menu_MultiPlayer_f ();
		mp_cursor = 1;
		break;
	case K_ENTER:
		{
			S_LocalSound ("misc/menu2.wav");
			switch (browser_selected)
			{
			case 0:
				if (browser_looking)
				{
					browser_looking = false;
				}
				else
				{
					char data[2];
					memset(g_browseServers, 0, sizeof(g_browseServers));
					memset(g_sortedServers, 0, sizeof(g_sortedServers));
					lastBrowseThink = 0.0;
					browser_numsv = 0;
					browser_numsorted = 0;
					browser_numqueried = 0;
					browser_numreplied = 0;
					browser_svnumbase = 0;
					browser_looking = true;

					strcpy(browser_status, "Waiting for master list.");

					//send a request to the qr master
					data[0] = C2M_GETLIST;
					data[1] = 0;
					NET_SendPacket (2, data, cl_master_adr);
				}
				break;
			case 1:
				if (browser_sortby == BROWSER_SORT_PING)
				{
					browser_sortby = BROWSER_SORT_PLAYERS;
				}
				else
				{
					browser_sortby = BROWSER_SORT_PING;
				}
				if (browser_numsorted)
				{ //re-sort
					qsort(g_sortedServers, browser_numsorted, sizeof(g_sortedServers[0]), M_ServerSort);
				}
				break;
			default:
				{
					int servSel = (browser_selected-NUM_BROWSER_ITEMS)+browser_svnumbase;
					if (servSel >= 0 && servSel < browser_numsorted)
					{
						char str[512];
						browseSv_t *sv = g_sortedServers[servSel];
						sprintf(str, "connect \"%i.%i.%i.%i:%i\"\n", sv->adr.ip[0],
							sv->adr.ip[1], sv->adr.ip[2], sv->adr.ip[3], BigShort(sv->adr.port));
						Cbuf_AddText(str);
						lastBrowseThink = 0.0;
						browser_numsv = 0;
						browser_numsorted = 0;
						browser_numqueried = 0;
						browser_numreplied = 0;
						browser_looking = false;

						key_dest = key_game;
						m_state = m_none;
						cls.demonum = m_save_demonum;
						if (cls.demonum != -1 && !cls.demoplayback && cls.state == ca_disconnected)
							CL_NextDemo ();
					}
				}
				break;
			}
		}
		break;
	case K_UPARROW:
		S_LocalSound ("misc/menu1.wav");
		browser_selected--;
		if (browser_selected < 0)
			browser_selected = numBrowserItems-1;
		break;

	case K_DOWNARROW:
		S_LocalSound ("misc/menu1.wav");
		browser_selected++;
		if (browser_selected >= numBrowserItems)
			browser_selected = 0;
		break;

	case K_LEFTARROW:
		S_LocalSound ("misc/menu1.wav");
		browser_svnumbase -= MAX_BROWSER_SV_DISPLAY;
		if (browser_svnumbase < 0)
		{
			browser_svnumbase = 0;
		}
		break;
	case K_RIGHTARROW:
		S_LocalSound ("misc/menu1.wav");
		browser_svnumbase += MAX_BROWSER_SV_DISPLAY;
		if ((browser_svnumbase+MAX_BROWSER_SV_DISPLAY) >= browser_numsorted)
		{
			browser_svnumbase = browser_numsorted-MAX_BROWSER_SV_DISPLAY;
			if (browser_svnumbase < 0)
			{
				browser_svnumbase = 0;
			}
		}
		break;
	default:
		break;
	}
}

//rww - ported back from normal quake
int		setup_cursor = 4;
int		setup_cursor_table[] = {40, 56, 80, 104, 140};

char	setup_hostname[16];
char	setup_myname[16];
int		setup_oldtop;
int		setup_oldbottom;
int		setup_top;
int		setup_bottom;

#define	NUM_SETUP_CMDS	5

void M_Menu_Setup_f (void)
{
	key_dest = key_menu;
	m_state = m_setup;
	m_entersound = true;
	Q_strcpy(setup_myname, name.string);
	Q_strcpy(setup_hostname, "N/A");
	setup_top = setup_oldtop = atoi(Info_ValueForKey (cls.userinfo, "topcolor"));//((int)cl_color.value) >> 4;
	setup_bottom = setup_oldbottom = atoi(Info_ValueForKey (cls.userinfo, "bottomcolor"));//((int)cl_color.value) & 15;
}


void M_Setup_Draw (void)
{
	qpic_t	*p;

	if (hr_plaque)
	{
		M_DrawHiResTex(0, 4, hr_plaque, 48, 216);
	}
	else
	{
		M_DrawTransPic (16, 4, Draw_CachePic ("gfx/qplaque.lmp") );
	}

	//p = Draw_CachePic ("gfx/p_multi.lmp");
	//M_DrawPic ( (320-p->width)/2, 4, p);
	M_Print(72, 8, "Multiplayer Setup");

	/*
	M_Print (64, 40, "Hostname");
	M_DrawTextBox (160, 32, 16, 1);
	M_Print (168, 40, setup_hostname);
	*/

	M_Print (64, 56, "Your name");
	M_DrawTextBox (160, 48, 16, 1);
	M_Print (168, 56, setup_myname);

	M_Print (64, 80, "Shirt color");
	M_Print (64, 104, "Pants color");

	M_DrawTextBox (64, 140-8, 14, 1);
	M_Print (72, 140, "Accept Changes");

	p = Draw_CachePic ("gfx/bigbox.lmp");
	M_DrawTransPic (160, 64, p);
	p = Draw_CachePic ("gfx/menuplyr.lmp");
	M_BuildTranslationTable(setup_top*16, setup_bottom*16);
	M_DrawTransPicTranslate (172, 72, p);

	M_DrawCharacter (56, setup_cursor_table [setup_cursor], 12+((int)(realtime*4)&1));

	/*
	if (setup_cursor == 0)
		M_DrawCharacter (168 + 8*strlen(setup_hostname), setup_cursor_table [setup_cursor], 10+((int)(realtime*4)&1));
		*/

	if (setup_cursor == 1)
		M_DrawCharacter (168 + 8*strlen(setup_myname), setup_cursor_table [setup_cursor], 10+((int)(realtime*4)&1));
}


void M_Setup_Key (int k)
{
	int			l;

	switch (k)
	{
	case K_ESCAPE:
		S_LocalSound ("misc/menu2.wav");
		M_Menu_MultiPlayer_f ();
		break;

	case K_UPARROW:
		S_LocalSound ("misc/menu1.wav");
		setup_cursor--;
		if (setup_cursor < 1)
			setup_cursor = NUM_SETUP_CMDS-1;
		break;

	case K_DOWNARROW:
		S_LocalSound ("misc/menu1.wav");
		setup_cursor++;
		if (setup_cursor >= NUM_SETUP_CMDS)
			setup_cursor = 1;
		break;

	case K_LEFTARROW:
		if (setup_cursor < 2)
			return;
		S_LocalSound ("misc/menu3.wav");
		if (setup_cursor == 2)
			setup_top = setup_top - 1;
		if (setup_cursor == 3)
			setup_bottom = setup_bottom - 1;
		break;
	case K_RIGHTARROW:
		if (setup_cursor < 2)
			return;
forward:
		S_LocalSound ("misc/menu3.wav");
		if (setup_cursor == 2)
			setup_top = setup_top + 1;
		if (setup_cursor == 3)
			setup_bottom = setup_bottom + 1;
		break;

	case K_ENTER:
		if (setup_cursor == 0 || setup_cursor == 1)
			return;

		if (setup_cursor == 2 || setup_cursor == 3)
			goto forward;

		// setup_cursor == 4 (OK)
		if (Q_strcmp(name.string, setup_myname) != 0)
			Cbuf_AddText ( va ("name \"%s\"\n", setup_myname) );
		/*
		if (Q_strcmp(hostname.string, setup_hostname) != 0)
			Cvar_Set("hostname", setup_hostname);
			*/
		if (setup_top != setup_oldtop || setup_bottom != setup_oldbottom)
			Cbuf_AddText( va ("color %i %i\n", setup_top, setup_bottom) );
		m_entersound = true;
		M_Menu_MultiPlayer_f ();
		break;

	case K_BACKSPACE:
		/*
		if (setup_cursor == 0)
		{
			if (strlen(setup_hostname))
				setup_hostname[strlen(setup_hostname)-1] = 0;
		}
		*/

		if (setup_cursor == 1)
		{
			if (strlen(setup_myname))
				setup_myname[strlen(setup_myname)-1] = 0;
		}
		break;

	default:
		if (k < 32 || k > 127)
			break;
		if (setup_cursor == 0)
		{
			l = strlen(setup_hostname);
			if (l < 15)
			{
				setup_hostname[l+1] = 0;
				setup_hostname[l] = k;
			}
		}
		if (setup_cursor == 1)
		{
			l = strlen(setup_myname);
			if (l < 15)
			{
				setup_myname[l+1] = 0;
				setup_myname[l] = k;
			}
		}
		break;
	}

	if (setup_top > 13)
		setup_top = 0;
	if (setup_top < 0)
		setup_top = 13;
	if (setup_bottom > 13)
		setup_bottom = 0;
	if (setup_bottom < 0)
		setup_bottom = 13;
}

//=============================================================================
/* OPTIONS MENU */

#define	OPTIONS_ITEMS	18 //rww - was 16

#define	SLIDER_RANGE	10

int		options_cursor;

void M_Menu_Options_f (void)
{
	key_dest = key_menu;
	m_state = m_options;
	m_entersound = true;
}


void M_AdjustSliders (int dir)
{
	S_LocalSound ("misc/menu3.wav");

	switch (options_cursor)
	{
		/*
	case 3:	// screen size
		scr_viewsize.value += dir * 10;
		if (scr_viewsize.value < 30)
			scr_viewsize.value = 30;
		if (scr_viewsize.value > 120)
			scr_viewsize.value = 120;
		Cvar_SetValue ("viewsize", scr_viewsize.value);
		break;
		*/
	case 4:	// gamma
		v_gamma.value -= dir * 0.05;
		if (v_gamma.value < 0.5)
			v_gamma.value = 0.5;
		if (v_gamma.value > 1)
			v_gamma.value = 1;
		Cvar_SetValue ("gamma", v_gamma.value);
		break;
	case 5:	// mouse speed
		sensitivity.value += dir * 0.5;
		if (sensitivity.value < 1)
			sensitivity.value = 1;
		if (sensitivity.value > 11)
			sensitivity.value = 11;
		Cvar_SetValue ("sensitivity", sensitivity.value);
		break;
	case 6:	// music volume
#ifdef _WIN32
		bgmvolume.value += dir * 1.0;
#else
		bgmvolume.value += dir * 0.1;
#endif
		if (bgmvolume.value < 0)
			bgmvolume.value = 0;
		if (bgmvolume.value > 1)
			bgmvolume.value = 1;
		Cvar_SetValue ("bgmvolume", bgmvolume.value);
		break;
	case 7:	// sfx volume
		volume.value += dir * 0.1;
		if (volume.value < 0)
			volume.value = 0;
		if (volume.value > 1)
			volume.value = 1;
		Cvar_SetValue ("volume", volume.value);
		break;
		
	case 8:	// allways run
		if (cl_forwardspeed.value > 200)
		{
			Cvar_SetValue ("cl_forwardspeed", 200);
			Cvar_SetValue ("cl_backspeed", 200);
		}
		else
		{
			Cvar_SetValue ("cl_forwardspeed", 400);
			Cvar_SetValue ("cl_backspeed", 400);
		}
		break;
	
	case 9:	// invert mouse
		Cvar_SetValue ("m_pitch", -m_pitch.value);
		break;
	
	case 10:	// lookspring
		Cvar_SetValue ("lookspring", !lookspring.value);
		break;
	
	case 11:	// lookstrafe
		Cvar_SetValue ("lookstrafe", !lookstrafe.value);
		break;

	case 12:
		Cvar_SetValue ("cl_sbar", !cl_sbar.value);
		break;

	case 13:
		Cvar_SetValue ("cl_hudswap", !cl_hudswap.value);
		break; //rww - added break

	case 16:	// _windowed_mouse //rww - was 15
		Cvar_SetValue ("_windowed_mouse", !_windowed_mouse.value);
		break;
	case 17:	//rww - cl_alwaysmlook
		Cvar_SetValue ("cl_alwaysmlook", !cl_alwaysmlook.value);
		break;
	}
}


void M_DrawSlider (int x, int y, float range)
{
	int	i;

	if (range < 0)
		range = 0;
	if (range > 1)
		range = 1;
	M_DrawCharacter (x-8, y, 128);
	for (i=0 ; i<SLIDER_RANGE ; i++)
		M_DrawCharacter (x + i*8, y, 129);
	M_DrawCharacter (x+i*8, y, 130);
	M_DrawCharacter (x + (SLIDER_RANGE-1)*8 * range, y, 131);
}

void M_DrawCheckbox (int x, int y, int on)
{
#if 0
	if (on)
		M_DrawCharacter (x, y, 131);
	else
		M_DrawCharacter (x, y, 129);
#endif
	if (on)
		M_Print (x, y, "on");
	else
		M_Print (x, y, "off");
}

void M_Options_Draw (void)
{
	float		r;
	qpic_t	*p;
	if (hr_plaque)
	{ //rww
		M_DrawHiResTex(0, 4, hr_plaque, 48, 216);
	}
	else
	{
		M_DrawTransPic (0, 4, Draw_CachePic ("gfx/qplaque.lmp") );
	}
	//p = Draw_CachePic ("gfx/p_option.lmp");
	//M_DrawPic ( (320-p->width)/2, 4, p);
	M_Print(72, 8, "Options"); //rww
	
	M_Print (16, 32, "    Customize controls");
	M_Print (220, 32, "More..."); //rww
	M_Print (16, 40, "         Go to console");
	M_Print (220, 40, "ok"); //rww
	M_Print (16, 48, "     Reset to defaults");
	M_Print (220, 48, "ok"); //rww

	/*
	M_Print (16, 56, "           Screen size");
	r = (scr_viewsize.value - 30) / (120 - 30);
	M_DrawSlider (220, 56, r);
	*/

	M_Print (16, 64, "                 Gamma"); //rww - changed from Brightness
	r = (1.0 - v_gamma.value) / 0.5;
	M_DrawSlider (220, 64, r);

	M_Print (16, 72, "           Mouse Speed");
	r = (sensitivity.value - 1)/10;
	M_DrawSlider (220, 72, r);

	M_Print (16, 80, "       CD Music Volume");
	r = bgmvolume.value;
	M_DrawSlider (220, 80, r);

	M_Print (16, 88, "          Sound Volume");
	r = volume.value;
	M_DrawSlider (220, 88, r);

	M_Print (16, 96,  "            Always Run");
	M_DrawCheckbox (220, 96, cl_forwardspeed.value > 200);

	M_Print (16, 104, "          Invert Mouse");
	M_DrawCheckbox (220, 104, m_pitch.value < 0);

	M_Print (16, 112, "            Lookspring");
	M_DrawCheckbox (220, 112, lookspring.value);

	M_Print (16, 120, "            Lookstrafe");
	M_DrawCheckbox (220, 120, lookstrafe.value);

	M_Print (16, 128, "    Use old status bar");
	M_DrawCheckbox (220, 128, cl_sbar.value);

	M_Print (16, 136, "      HUD on left side");
	M_DrawCheckbox (220, 136, cl_hudswap.value);

	if (vid_menudrawfn)
	{ //rww - changed y, added more...
		M_Print (16, 144, "         Video Modes");
		M_Print (220, 144, "More...");
	}

	//rww begin
	if (vid_optmenudrawfn)
	{
		M_Print (16, 152, "         Video Options");
		M_Print (220, 152, "More...");
	}
	//rww end

#ifdef _WIN32
	if (modestate == MS_WINDOWED)
	{
#endif
		M_Print (16, 160, "             Use Mouse"); //rww - 152 -> 160
		M_DrawCheckbox (220, 160, _windowed_mouse.value); //rww - 152 -> 160
#ifdef _WIN32
	}
#endif

	//rww
	M_Print (16, 168, "      Always Mouselook");
	M_DrawCheckbox (220, 168, cl_alwaysmlook.value);

// cursor
	M_DrawCharacter (200, 32 + options_cursor*8, 12+((int)(realtime*4)&1));
}


void M_Options_Key (int k)
{
	switch (k)
	{
	case K_ESCAPE:
		M_Menu_Main_f ();
		break;
		
	case K_ENTER:
		m_entersound = true;
		switch (options_cursor)
		{
		case 0:
			M_Menu_Keys_f ();
			break;
		case 1:
			m_state = m_none;
			Con_ToggleConsole_f ();
			break;
		case 2:
			Cbuf_AddText ("exec default.cfg\n");
			break;
		case 14:
			M_Menu_Video_f ();
			break;
		case 15: //rww
			M_Menu_OptVideo_f ();
			break;
		default:
			M_AdjustSliders (1);
			break;
		}
		return;
	
	case K_UPARROW:
		S_LocalSound ("misc/menu1.wav");
		options_cursor--;
		if (options_cursor == 3)
		{
			options_cursor = 2;
		}
		if (options_cursor < 0)
			options_cursor = OPTIONS_ITEMS-1;
		break;

	case K_DOWNARROW:
		S_LocalSound ("misc/menu1.wav");
		options_cursor++;
		if (options_cursor == 3)
		{
			options_cursor = 4;
		}
		if (options_cursor >= OPTIONS_ITEMS)
			options_cursor = 0;
		break;	

	case K_LEFTARROW:
		M_AdjustSliders (-1);
		break;

	case K_RIGHTARROW:
		M_AdjustSliders (1);
		break;
	}

	if (options_cursor == 14 && vid_menudrawfn == NULL)
	{
		if (k == K_UPARROW)
			options_cursor = 13;
		else
			options_cursor = 0;
	}

	if ((options_cursor == 16) 
#ifdef _WIN32
	&& (modestate != MS_WINDOWED)
#endif
	)
	{
		if (k == K_UPARROW)
			options_cursor = 15;//14; //rww
		else
			options_cursor = 17;//0; //rww
	}
}


//=============================================================================
/* KEYS MENU */

char *bindnames[][2] =
{
{"+attack", 		"attack"},
{"impulse 10", 		"change weapon"},
{"impulse 12", 		"last weapon"},
{"+jump", 			"jump / swim up"},
{"+forward", 		"walk forward"},
{"+back", 			"backpedal"},
{"+left", 			"turn left"},
{"+right", 			"turn right"},
{"+speed", 			"run"},
{"+moveleft", 		"step left"},
{"+moveright", 		"step right"},
{"+strafe", 		"sidestep"},
{"+lookup", 		"look up"},
{"+lookdown", 		"look down"},
{"centerview", 		"center view"},
{"+mlook", 			"mouse look"},
{"+klook", 			"keyboard look"},
{"+moveup",			"swim up"},
{"+movedown",		"swim down"},
{"automaptoggle",	"automap"},
{"messagemode",		"chat"},
{"messagemode2",	"team chat"}
};

#define	NUMCOMMANDS	(sizeof(bindnames)/sizeof(bindnames[0]))

int		keys_cursor;
int		bind_grab;

//rww
qboolean M_KeyMenuActive(void)
{
	return m_state == m_keys;
}

void M_Menu_Keys_f (void)
{
	key_dest = key_menu;
	m_state = m_keys;
	m_entersound = true;
}


void M_FindKeysForCommand (char *command, int *twokeys)
{
	int		count;
	int		j;
	int		l;
	char	*b;

	twokeys[0] = twokeys[1] = -1;
	l = strlen(command);
	count = 0;

	for (j=0 ; j<256 ; j++)
	{
		b = keybindings[j];
		if (!b)
			continue;
		//if (!strncmp (b, command, l) )
		//rww - strncmp messes things up (messagemode/messagemode2), don't see why it was used
		if (!strcmp(b, command))
		{
			twokeys[count] = j;
			count++;
			if (count == 2)
				break;
		}
	}
}

void M_UnbindCommand (char *command)
{
	int		j;
	int		l;
	char	*b;

	l = strlen(command);

	for (j=0 ; j<256 ; j++)
	{
		b = keybindings[j];
		if (!b)
			continue;
		//if (!strncmp (b, command, l) )
		if (!strcmp (b, command) )
		//rww - strncmp crap again
			Key_SetBinding (j, "");
	}
}


void M_Keys_Draw (void)
{
	int		i, l;
	int		keys[2];
	char	*name;
	int		x, y;
	qpic_t	*p;

//	p = Draw_CachePic ("gfx/ttl_cstm.lmp");
//	M_DrawPic ( (320-p->width)/2, 4, p);
	M_Print(72, 8, "Customize controls"); //rww

	if (bind_grab)
		M_Print (12, 32, "Press a key or button for this action");
	else
		M_Print (18, 32, "Enter to change, backspace to clear");
		
// search for known bindings
	for (i=0 ; i<NUMCOMMANDS ; i++)
	{
		y = 48 + 8*i;

		M_Print (16, y, bindnames[i][1]);

		l = strlen (bindnames[i][0]);
		
		M_FindKeysForCommand (bindnames[i][0], keys);
		
		if (keys[0] == -1)
		{
			M_Print (140, y, "???");
		}
		else
		{
			name = Key_KeynumToString (keys[0]);
			M_Print (140, y, name);
			x = strlen(name) * 8;
			if (keys[1] != -1)
			{
				M_Print (140 + x + 8, y, "or");
				M_Print (140 + x + 32, y, Key_KeynumToString (keys[1]));
			}
		}
	}
	
	if (bind_grab)
		M_DrawCharacter (130, 48 + keys_cursor*8, '=');
	else
		M_DrawCharacter (130, 48 + keys_cursor*8, 12+((int)(realtime*4)&1));
}


void M_Keys_Key (int k)
{
	char	cmd[80];
	int		keys[2];
	
	if (bind_grab)
	{	// defining a key
		S_LocalSound ("misc/menu1.wav");
		if (k == K_ESCAPE)
		{
			bind_grab = false;
		}
		else if (k != '`')
		{
			sprintf (cmd, "bind %s \"%s\"\n", Key_KeynumToString (k), bindnames[keys_cursor][0]);			
			Cbuf_InsertText (cmd);
		}
		
		bind_grab = false;
		return;
	}
	
	switch (k)
	{
	case K_ESCAPE:
		M_Menu_Options_f ();
		break;

	case K_LEFTARROW:
	case K_UPARROW:
		S_LocalSound ("misc/menu1.wav");
		keys_cursor--;
		if (keys_cursor < 0)
			keys_cursor = NUMCOMMANDS-1;
		break;

	case K_DOWNARROW:
	case K_RIGHTARROW:
		S_LocalSound ("misc/menu1.wav");
		keys_cursor++;
		if (keys_cursor >= NUMCOMMANDS)
			keys_cursor = 0;
		break;

	case K_ENTER:		// go into bind mode
		M_FindKeysForCommand (bindnames[keys_cursor][0], keys);
		S_LocalSound ("misc/menu2.wav");
		if (keys[1] != -1)
			M_UnbindCommand (bindnames[keys_cursor][0]);
		bind_grab = true;
		break;

	case K_BACKSPACE:		// delete bindings
	case K_DEL:				// delete bindings
		S_LocalSound ("misc/menu2.wav");
		M_UnbindCommand (bindnames[keys_cursor][0]);
		break;
	}
}

//=============================================================================
/* VIDEO MENU */

void M_Menu_Video_f (void)
{
	key_dest = key_menu;
	m_state = m_video;
	m_entersound = true;
}

void M_Menu_OptVideo_f (void) //rww
{
	key_dest = key_menu;
	m_state = m_optvideo;
	m_entersound = true;
}


void M_Video_Draw (void)
{
	(*vid_menudrawfn) ();
}


void M_Video_Key (int key)
{
	(*vid_menukeyfn) (key);
}

//rww begin
void M_OptVideo_Draw (void)
{
	(*vid_optmenudrawfn) ();
}


void M_OptVideo_Key (int key)
{
	(*vid_optmenukeyfn) (key);
}
//rww end


//=============================================================================
/* HELP MENU */

int		help_page;
#define	NUM_HELP_PAGES	6


void M_Menu_Help_f (void)
{
	//rww
	/*
	key_dest = key_menu;
	m_state = m_help;
	m_entersound = true;
	help_page = 0;
	*/
}



void M_Help_Draw (void)
{
	M_DrawPic (0, 0, Draw_CachePic ( va("gfx/help%i.lmp", help_page)) );
}


void M_Help_Key (int key)
{
	switch (key)
	{
	case K_ESCAPE:
		M_Menu_Main_f ();
		break;
		
	case K_UPARROW:
	case K_RIGHTARROW:
		m_entersound = true;
		if (++help_page >= NUM_HELP_PAGES)
			help_page = 0;
		break;

	case K_DOWNARROW:
	case K_LEFTARROW:
		m_entersound = true;
		if (--help_page < 0)
			help_page = NUM_HELP_PAGES-1;
		break;
	}

}

//=============================================================================
/* QUIT MENU */

int		msgNumber;
int		m_quit_prevstate;
qboolean	wasInMenus;

char *quitMessage [] = 
{
/* .........1.........2.... */
  "  Are you gonna quit    ",
  "  this game just like   ",
  "   everything else?     ",
  "                        ",
 
  " Milord, methinks that  ",
  "   thou art a lowly     ",
  " quitter. Is this true? ",
  "                        ",

  " Do I need to bust your ",
  "  face open for trying  ",
  "        to quit?        ",
  "                        ",

  " Man, I oughta smack you",
  "   for trying to quit!  ",
  "     Press Y to get     ",
  "      smacked out.      ",
 
  " Press Y to quit like a ",
  "   big loser in life.   ",
  "  Press N to stay proud ",
  "    and successful!     ",
 
  "   If you press Y to    ",
  "  quit, I will summon   ",
  "  Satan all over your   ",
  "      hard drive!       ",
 
  "  Um, Asmodeus dislikes ",
  " his children trying to ",
  " quit. Press Y to return",
  "   to your Tinkertoys.  ",
 
  "  If you quit now, I'll ",
  "  throw a blanket-party ",
  "   for you next time!   ",
  "                        "
};

void M_Menu_Quit_f (void)
{
	/*
	if (m_state == m_quit)
		return;
	wasInMenus = (key_dest == key_menu);
	key_dest = key_menu;
	m_quit_prevstate = m_state;
	m_state = m_quit;
	m_entersound = true;
	msgNumber = rand()&7;
	*/
	//rww
	key_dest = key_console;
	CL_Disconnect ();
	Sys_Quit ();
}


void M_Quit_Key (int key)
{
	switch (key)
	{
	case K_ESCAPE:
	case 'n':
	case 'N':
		if (wasInMenus)
		{
			m_state = m_quit_prevstate;
			m_entersound = true;
		}
		else
		{
			key_dest = key_game;
			m_state = m_none;
		}
		break;

	case 'Y':
	case 'y':
		key_dest = key_console;
		CL_Disconnect ();
		Sys_Quit ();
		break;

	default:
		break;
	}

}

void M_Menu_SinglePlayer_f (void) {
	m_state = m_singleplayer;
}

void M_SinglePlayer_Draw (void) {
	qpic_t	*p;

	if (hr_plaque)
	{ //rww
		M_DrawHiResTex(0, 4, hr_plaque, 48, 216);
	}
	else
	{
		M_DrawTransPic (0, 4, Draw_CachePic ("gfx/qplaque.lmp") );
	}
//	M_DrawTransPic (16, 4, Draw_CachePic ("gfx/qplaque.lmp") );
	p = Draw_CachePic ("gfx/ttl_sgl.lmp");
	M_DrawPic ( (320-p->width)/2, 4, p);
//	M_DrawTransPic (72, 32, Draw_CachePic ("gfx/sp_menu.lmp") );

	M_DrawTextBox (60, 10*8, 23, 4);	
	M_PrintWhite (92, 12*8, "QuakeWorld is for");
	M_PrintWhite (88, 13*8, "Internet play only");

}

void M_SinglePlayer_Key (key) {
	if (key == K_ESCAPE || key==K_ENTER)
		m_state = m_main;
}

#define NUM_MP_MENU_ITEMS	3

char mp_ipconnect[64];

void M_Menu_MultiPlayer_f (void) {
	mp_cursor = 0;
	strcpy(mp_ipconnect, "localhost");
	m_state = m_multiplayer;
}

//rww - new menu
void M_MultiPlayer_Draw (void) {
	qpic_t	*p;

	if (hr_plaque)
	{ //rww
		M_DrawHiResTex(0, 4, hr_plaque, 48, 216);
	}
	else
	{
		M_DrawTransPic (0, 4, Draw_CachePic ("gfx/qplaque.lmp") );
	}
//	M_DrawTransPic (16, 4, Draw_CachePic ("gfx/qplaque.lmp") );
	//p = Draw_CachePic ("gfx/p_multi.lmp");
	//M_DrawPic ( (320-p->width)/2, 4, p);
	M_Print(72, 8, "Multiplayer");

	M_Print(96, 32, "Setup");
	M_Print(96, 40, "Server Browser");
	M_Print(96, 48, "IP Connect:");
	M_Print(192, 48, mp_ipconnect);
	if (cls.state != ca_disconnected && !cls.demoplayback)
	{
		M_Print(96, 56, "Disconnect");
	}
	else if (mp_cursor == 3)
	{
		mp_cursor = 0;
	}

//	M_DrawTransPic (72, 32, Draw_CachePic ("gfx/sp_menu.lmp") );

	M_DrawCharacter (80, 32 + (mp_cursor*8), 12+((int)(realtime*4)&1));

	if (mp_cursor == 2)
	{
		M_DrawCharacter (192 + 8*strlen(mp_ipconnect), 48, 10+((int)(realtime*4)&1));
	}

	/*
	M_DrawTextBox (46, 8*8, 27, 9);	
	M_PrintWhite (72, 10*8, "If you want to find QW  ");
	M_PrintWhite (72, 11*8, "games, head on over to: ");
	     M_Print (72, 12*8, "   www.quakeworld.net   ");
	M_PrintWhite (72, 13*8, "          or            ");
	     M_Print (72, 14*8, "   www.quakespy.com     ");
	M_PrintWhite (72, 15*8, "For pointers on getting ");
	M_PrintWhite (72, 16*8, "        started!        ");
	*/
}

void M_MultiPlayer_Key (key) {
	int numMenuItems = NUM_MP_MENU_ITEMS;
	if (cls.state != ca_disconnected && !cls.demoplayback)
	{ //disconnect becomes available
		numMenuItems++;
	}

	switch (key)
	{
	case K_ESCAPE:
		S_LocalSound ("misc/menu2.wav");
		m_state = m_main;
		break;

	case K_LEFTARROW:
	case K_UPARROW:
		S_LocalSound ("misc/menu1.wav");
		mp_cursor--;
		if (mp_cursor < 0)
			mp_cursor = numMenuItems-1;
		break;

	case K_DOWNARROW:
	case K_RIGHTARROW:
		S_LocalSound ("misc/menu1.wav");
		mp_cursor++;
		if (mp_cursor >= numMenuItems)
			mp_cursor = 0;
		break;

	case K_ENTER:
		S_LocalSound ("misc/menu2.wav");
		switch (mp_cursor)
		{
		case 0:
			M_Menu_Setup_f();
			break;
		case 1:
			M_Menu_SvBrowser_f();
			break;
		case 2:
			{
				char str[256];
				sprintf(str, "connect \"%s\"\n", mp_ipconnect);
				key_dest = key_game;
				m_state = m_none;
				cls.demonum = m_save_demonum;
				Cbuf_AddText(str);
			}
			break;
		case 3:
			Cbuf_AddText("disconnect\n");
			break;
		default:
			break;
		}
		break;

	case K_BACKSPACE:
		if (mp_cursor == 2)
		{
			if (strlen(mp_ipconnect))
				mp_ipconnect[strlen(mp_ipconnect)-1] = 0;
		}
		break;

	default:
		if (key < 32 || key > 127)
			break;
		if (mp_cursor == 2)
		{
			int l = (int)strlen(mp_ipconnect);
			if (l < 35)
			{
				mp_ipconnect[l+1] = 0;
				mp_ipconnect[l] = key;
			}
		}
		break;
	}
}

void M_Quit_Draw (void)
{
#define VSTR(x) #x
#define VSTR2(x) VSTR(x)
	char *cmsg[] = {
//    0123456789012345678901234567890123456789
	"0            QuakeWorld",
	"1    version " VSTR2(VERSION) " by id Software",
	"0Programming",
	"1 John Carmack    Michael Abrash",
	"1 John Cash       Christian Antkow",
	"0Additional Programming",
	"1 Dave 'Zoid' Kirsch",
	"1 Jack 'morbid' Mathews",
	"0Id Software is not responsible for",
    "0providing technical support for",
	"0QUAKEWORLD(tm). (c)1996 Id Software,",
	"0Inc.  All Rights Reserved.",
	"0QUAKEWORLD(tm) is a trademark of Id",
	"0Software, Inc.",
	"1NOTICE: THE COPYRIGHT AND TRADEMARK",
	"1NOTICES APPEARING  IN YOUR COPY OF",
	"1QUAKE(r) ARE NOT MODIFIED BY THE USE",
	"1OF QUAKEWORLD(tm) AND REMAIN IN FULL",
	"1FORCE.",
	"0NIN(r) is a registered trademark",
	"0licensed to Nothing Interactive, Inc.",
	"0All rights reserved. Press y to exit",
	NULL };
	char **p;
	int y;

	if (wasInMenus)
	{
		m_state = m_quit_prevstate;
		m_recursiveDraw = true;
		M_Draw ();
		m_state = m_quit;
	}
#if 1
	M_DrawTextBox (0, 0, 38, 23);
	y = 12;
	for (p = cmsg; *p; p++, y += 8) {
		if (**p == '0')
			M_PrintWhite (16, y, *p + 1);
		else
			M_Print (16, y,	*p + 1);
	}
#else
	M_DrawTextBox (56, 76, 24, 4);
	M_Print (64, 84,  quitMessage[msgNumber*4+0]);
	M_Print (64, 92,  quitMessage[msgNumber*4+1]);
	M_Print (64, 100, quitMessage[msgNumber*4+2]);
	M_Print (64, 108, quitMessage[msgNumber*4+3]);
#endif
}



//=============================================================================
/* Menu Subsystem */


void M_Init (void)
{
	Cmd_AddCommand ("togglemenu", M_ToggleMenu_f);

	Cmd_AddCommand ("menu_main", M_Menu_Main_f);
	Cmd_AddCommand ("menu_options", M_Menu_Options_f);
	Cmd_AddCommand ("menu_keys", M_Menu_Keys_f);
	Cmd_AddCommand ("menu_video", M_Menu_Video_f);
	Cmd_AddCommand ("menu_optvideo", M_Menu_OptVideo_f);
	Cmd_AddCommand ("help", M_Menu_Help_f);
	Cmd_AddCommand ("menu_quit", M_Menu_Quit_f);
}


void M_Draw (void)
{
	if (m_state == m_none || key_dest != key_menu)
		return;

	if (!m_recursiveDraw)
	{
		scr_copyeverything = 1;

		if (scr_con_current)
		{
			Draw_ConsoleBackground (vid.height);
			VID_UnlockBuffer ();
			S_ExtraUpdate ();
			VID_LockBuffer ();
		}
		else
			Draw_FadeScreen ();

		scr_fullupdate = 0;
	}
	else
	{
		m_recursiveDraw = false;
	}

	switch (m_state)
	{
	case m_none:
		break;

	case m_main:
		M_Main_Draw ();
		break;

	case m_singleplayer:
		M_SinglePlayer_Draw ();
		break;

	case m_load:
//		M_Load_Draw ();
		break;

	case m_save:
//		M_Save_Draw ();
		break;

	case m_multiplayer:
		M_MultiPlayer_Draw ();
		break;

	case m_infomenu: //rww
		M_InfoMenu_Draw();
		break;

	case m_setup:
		M_Setup_Draw ();
		break;

	case m_svbrowser: //rww
		M_SvBrowser_Draw ();
		break;

	case m_net:
//		M_Net_Draw ();
		break;

	case m_options:
		M_Options_Draw ();
		break;

	case m_keys:
		M_Keys_Draw ();
		break;

	case m_video:
		M_Video_Draw ();
		break;

	case m_optvideo: //rww
		M_OptVideo_Draw ();
		break;

	case m_help:
		M_Help_Draw ();
		break;

	case m_quit:
		M_Quit_Draw ();
		break;

	case m_serialconfig:
//		M_SerialConfig_Draw ();
		break;

	case m_modemconfig:
//		M_ModemConfig_Draw ();
		break;

	case m_lanconfig:
//		M_LanConfig_Draw ();
		break;

	case m_gameoptions:
//		M_GameOptions_Draw ();
		break;

	case m_search:
//		M_Search_Draw ();
		break;

	case m_slist:
//		M_ServerList_Draw ();
		break;
	}

	if (m_entersound)
	{
		S_LocalSound ("misc/menu2.wav");
		m_entersound = false;
	}

	VID_UnlockBuffer ();
	S_ExtraUpdate ();
	VID_LockBuffer ();
}


void M_Keydown (int key)
{
	switch (m_state)
	{
	case m_none:
		return;

	case m_main:
		M_Main_Key (key);
		return;

	case m_singleplayer:
		M_SinglePlayer_Key (key);
		return;

	case m_load:
//		M_Load_Key (key);
		return;

	case m_save:
//		M_Save_Key (key);
		return;

	case m_multiplayer:
		M_MultiPlayer_Key (key);
		return;

	case m_infomenu: //rww
		M_InfoMenu_Key(key);
		break;

	case m_setup:
		M_Setup_Key (key);
		return;

	case m_svbrowser:
		M_SvBrowser_Key (key);
		return;

	case m_net:
//		M_Net_Key (key);
		return;

	case m_options:
		M_Options_Key (key);
		return;

	case m_keys:
		M_Keys_Key (key);
		return;

	case m_video:
		M_Video_Key (key);
		return;

	case m_optvideo: //rww
		M_OptVideo_Key (key);
		return;

	case m_help:
		M_Help_Key (key);
		return;

	case m_quit:
		M_Quit_Key (key);
		return;

	case m_serialconfig:
//		M_SerialConfig_Key (key);
		return;

	case m_modemconfig:
//		M_ModemConfig_Key (key);
		return;

	case m_lanconfig:
//		M_LanConfig_Key (key);
		return;

	case m_gameoptions:
//		M_GameOptions_Key (key);
		return;

	case m_search:
//		M_Search_Key (key);
		break;

	case m_slist:
//		M_ServerList_Key (key);
		return;
	}
}


